from collections import defaultdict
from typing import Dict
from typing import List

import torch

from funasr_local.modules.rnn.attentions import AttAdd
from funasr_local.modules.rnn.attentions import AttCov
from funasr_local.modules.rnn.attentions import AttCovLoc
from funasr_local.modules.rnn.attentions import AttDot
from funasr_local.modules.rnn.attentions import AttForward
from funasr_local.modules.rnn.attentions import AttForwardTA
from funasr_local.modules.rnn.attentions import AttLoc
from funasr_local.modules.rnn.attentions import AttLoc2D
from funasr_local.modules.rnn.attentions import AttLocRec
from funasr_local.modules.rnn.attentions import AttMultiHeadAdd
from funasr_local.modules.rnn.attentions import AttMultiHeadDot
from funasr_local.modules.rnn.attentions import AttMultiHeadLoc
from funasr_local.modules.rnn.attentions import AttMultiHeadMultiResLoc
from funasr_local.modules.rnn.attentions import NoAtt
from funasr_local.modules.attention import MultiHeadedAttention


from funasr_local.train.abs_espnet_model import AbsESPnetModel


@torch.no_grad()
def calculate_all_attentions(
    model: AbsESPnetModel, batch: Dict[str, torch.Tensor]
) -> Dict[str, List[torch.Tensor]]:
    """Derive the outputs from the all attention layers

    Args:
        model:
        batch: same as forward
    Returns:
        return_dict: A dict of a list of tensor.
        key_names x batch x (D1, D2, ...)

    """
    bs = len(next(iter(batch.values())))
    assert all(len(v) == bs for v in batch.values()), {
        k: v.shape for k, v in batch.items()
    }

    # 1. Register forward_hook fn to save the output from specific layers
    outputs = {}
    handles = {}
    for name, modu in model.named_modules():

        def hook(module, input, output, name=name):
            if isinstance(module, MultiHeadedAttention):
                # NOTE(kamo): MultiHeadedAttention doesn't return attention weight
                # attn: (B, Head, Tout, Tin)
                outputs[name] = module.attn.detach().cpu()
            elif isinstance(module, AttLoc2D):
                c, w = output
                # w: previous concate attentions
                # w: (B, nprev, Tin)
                att_w = w[:, -1].detach().cpu()
                outputs.setdefault(name, []).append(att_w)
            elif isinstance(module, (AttCov, AttCovLoc)):
                c, w = output
                assert isinstance(w, list), type(w)
                # w: list of previous attentions
                # w: nprev x (B, Tin)
                att_w = w[-1].detach().cpu()
                outputs.setdefault(name, []).append(att_w)
            elif isinstance(module, AttLocRec):
                # w: (B, Tin)
                c, (w, (att_h, att_c)) = output
                att_w = w.detach().cpu()
                outputs.setdefault(name, []).append(att_w)
            elif isinstance(
                module,
                (
                    AttMultiHeadDot,
                    AttMultiHeadAdd,
                    AttMultiHeadLoc,
                    AttMultiHeadMultiResLoc,
                ),
            ):
                c, w = output
                # w: nhead x (B, Tin)
                assert isinstance(w, list), type(w)
                att_w = [_w.detach().cpu() for _w in w]
                outputs.setdefault(name, []).append(att_w)
            elif isinstance(
                module,
                (
                    AttAdd,
                    AttDot,
                    AttForward,
                    AttForwardTA,
                    AttLoc,
                    NoAtt,
                ),
            ):
                c, w = output
                att_w = w.detach().cpu()
                outputs.setdefault(name, []).append(att_w)

        handle = modu.register_forward_hook(hook)
        handles[name] = handle

    # 2. Just forward one by one sample.
    # Batch-mode can't be used to keep requirements small for each models.
    keys = []
    for k in batch:
        if not k.endswith("_lengths"):
            keys.append(k)

    return_dict = defaultdict(list)
    for ibatch in range(bs):
        # *: (B, L, ...) -> (1, L2, ...)
        _sample = {
            k: batch[k][ibatch, None, : batch[k + "_lengths"][ibatch]]
            if k + "_lengths" in batch
            else batch[k][ibatch, None]
            for k in keys
        }

        # *_lengths: (B,) -> (1,)
        _sample.update(
            {
                k + "_lengths": batch[k + "_lengths"][ibatch, None]
                for k in keys
                if k + "_lengths" in batch
            }
        )
        model(**_sample)

        # Derive the attention results
        for name, output in outputs.items():
            if isinstance(output, list):
                if isinstance(output[0], list):
                    # output: nhead x (Tout, Tin)
                    output = torch.stack(
                        [
                            # Tout x (1, Tin) -> (Tout, Tin)
                            torch.cat([o[idx] for o in output], dim=0)
                            for idx in range(len(output[0]))
                        ],
                        dim=0,
                    )
                else:
                    # Tout x (1, Tin) -> (Tout, Tin)
                    output = torch.cat(output, dim=0)
            else:
                # output: (1, NHead, Tout, Tin) -> (NHead, Tout, Tin)
                output = output.squeeze(0)
            # output: (Tout, Tin) or (NHead, Tout, Tin)
            return_dict[name].append(output)
        outputs.clear()

    # 3. Remove all hooks
    for _, handle in handles.items():
        handle.remove()

    return dict(return_dict)



import os
import argparse
from multiprocessing import Pool
# from GOT.eval.merge_results import merge_outputs
# from GOT.eval.doctextVQA import doc_text_eval


def run_eval(chunk_id, model_name, gtfile_path, image_path, out_path, num_chunks, datatype, temperature):
    os.system("CUDA_VISIBLE_DEVICES=" + str(chunk_id) + " "
              + "python3 -m GOT.eval.eval_GOT_ocr" + " "
              + "--model-name" + " " + model_name + " "
              + "--gtfile_path" +  " " + gtfile_path + " "
              + "--image_path" + " " + image_path + " "
              + "--out_path" +  " " + out_path + " "
              + "--num-chunks" + " " +  str(num_chunks) + " "
              + "--chunk-idx" + " " +  str(chunk_id) + " "
              + "--temperature" + " " +  str(temperature) + " "
              + "--datatype" + " " + datatype
              )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--model-name", type=str, default="facebook/opt-350m")
    parser.add_argument("--gtfile_path", type=str, required=True)
    parser.add_argument("--image_path", type=str, required=True)
    parser.add_argument("--out_path", type=str, required=True)
    parser.add_argument("--num-chunks", type=int, default=1)
    parser.add_argument("--temperature", type=float, default=0.2)
    parser.add_argument("--datatype", type=str, required=True)  # Text or Doc
    # parser.add_argument("--eval", type=str, required=True)
    args = parser.parse_args()

    num_chunks = args.num_chunks
    
    if os.path.exists(args.out_path) == False:
        os.makedirs(args.out_path)


    with Pool(num_chunks) as p:
        for i in range(num_chunks):
            chunk_id = i
            p.apply_async(run_eval, (chunk_id, args.model_name, args.gtfile_path,
                                     args.image_path, args.out_path, num_chunks, args.datatype, args.temperature))
        p.close()
        p.join()




import requests
import httpx

from a2a.types import AgentCard
from a2a.utils.constants import AGENT_CARD_WELL_KNOWN_PATH


def get_agent_card(remote_agent_address: str) -> AgentCard:
    """Get the agent card."""
    if not remote_agent_address.startswith(("http://", "https://")):
        remote_agent_address = "http://" + remote_agent_address
    agent_card = requests.get(f"{remote_agent_address}{AGENT_CARD_WELL_KNOWN_PATH}")
    return AgentCard(**agent_card.json())


async def async_get_agent_card(
    remote_agent_address: str, http_client: httpx.AsyncClient
) -> AgentCard:
    """Get the agent card."""
    if not remote_agent_address.startswith(("http://", "https://")):
        remote_agent_address = "http://" + remote_agent_address
    agent_card = await http_client.get(f"{remote_agent_address}{AGENT_CARD_WELL_KNOWN_PATH}")
    return AgentCard(**agent_card.json())



import asyncio
import logging
from typing import List

from apipeline.frames.control_frames import StartFrame, EndFrame
from apipeline.frames.sys_frames import Frame, CancelFrame
from apipeline.processors.frame_processor import FrameDirection
from apipeline.processors.async_frame_processor import AsyncFrameProcessor

from src.types.frames.control_frames import IntervalFrame


class IntervalProcessor(AsyncFrameProcessor):
    """window.setInterval for IntervalFrame"""

    def __init__(
        self,
        *,
        interval_time_ms: int,
        types: List[type] = [],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.interval_time_ms = interval_time_ms
        self.types = types
        self._interval_task = None

    async def start(self):
        self._interval_task = self.get_event_loop().create_task(self._interval_handle())

    async def end(self):
        self._interval_task.cancel()
        await self._interval_task

    async def cancel(self):
        self._interval_task.cancel()
        await self._interval_task

    async def _interval_handle(self):
        while True:
            try:
                await asyncio.sleep(self.interval_time_ms / 1000)
                await self.push_frame(
                    IntervalFrame(
                        interval_time_ms=self.interval_time_ms,
                        types=self.types,
                    )
                )
            except Exception as e:
                logging.error(f"IntervalProcessor error: {e}")
            except asyncio.CancelledError:
                logging.info("IntervalProcessor cancelled")
                break

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        await super().process_frame(frame, direction)

        if isinstance(frame, StartFrame):
            await self.push_frame(frame, direction)
            await self.start()
        elif isinstance(frame, EndFrame):
            await self.push_frame(frame, direction)
            await self.end()
        elif isinstance(frame, CancelFrame):
            await self.cancel()
            await self.push_frame(frame, direction)
        else:
            await self.queue_frame(frame, direction)



import asyncio
from typing import Literal

import asyncclick as click
import colorama
from .agent import Agent


@click.command()
@click.option("--host", "host", default="localhost")
@click.option("--port", "port", default=6666)
@click.option("--mode", "mode", default="streaming")
@click.option("--question", "question", required=True)
async def a_main(
    host: str,
    port: int,
    mode: Literal["completion", "streaming"],
    question: str,
):
    """Main function to run the A2A Agent client.

    Args:
        host (str): The host address to run the server on.
        port (int): The port number to run the server on.
        mode (Literal['completion', 'streaming']): The mode to run the server on.
        question (str): The question to ask the Agent.
    """  # noqa: E501
    agent = Agent(
        mode="stream",
        token_stream_callback=None,
        agent_urls=[f"http://{host}:{port}/"],
    )
    async for chunk in agent.stream(question):
        if chunk.startswith('<Agent name="'):
            print(colorama.Fore.CYAN + chunk, end="", flush=True)
        elif chunk.startswith("</Agent>"):
            print(colorama.Fore.RESET + chunk, end="", flush=True)
        else:
            print(chunk, end="", flush=True)


def main() -> None:
    """Main function to run the github Repo Agent client."""
    asyncio.run(a_main())


"""
python -m a2a_server.cmd.client --question "what's achatbot?"
"""
if __name__ == "__main__":
    main()



import asyncio
from typing import AsyncGenerator

from google.adk.agents import LiveRequestQueue
from google.adk.agents.llm_agent import Agent
from google.adk.tools.function_tool import FunctionTool
from google.genai import Client
from google.genai import types as genai_types


async def monitor_stock_price(stock_symbol: str) -> AsyncGenerator[str, None]:
    """该函数将以持续、流式和异步的方式监控给定股票代码的价格。"""
    print(f"Start monitor stock price for {stock_symbol}!")

    # 模拟股票价格变化。
    await asyncio.sleep(4)
    price_alert1 = f"the price for {stock_symbol} is 300"
    yield price_alert1
    print(price_alert1)

    await asyncio.sleep(4)
    price_alert1 = f"the price for {stock_symbol} is 400"
    yield price_alert1
    print(price_alert1)

    await asyncio.sleep(20)
    price_alert1 = f"the price for {stock_symbol} is 900"
    yield price_alert1
    print(price_alert1)

    await asyncio.sleep(20)
    price_alert1 = f"the price for {stock_symbol} is 500"
    yield price_alert1
    print(price_alert1)


# 对于视频流，`input_stream: LiveRequestQueue` 是 ADK 保留的关键参数，用于传递视频流。
async def monitor_video_stream(
    input_stream: LiveRequestQueue,
) -> AsyncGenerator[str, None]:
    """监控视频流中有多少人。"""
    print("start monitor_video_stream!")
    client = Client(vertexai=False)
    prompt_text = "Count the number of people in this image. Just respond with a numeric number."
    last_count = None
    while True:
        last_valid_req = None
        print("Start monitoring loop")

        # 用此循环拉取最新图片并丢弃旧图片
        while not input_stream._queue.empty():
            live_req = await input_stream.get()

            if live_req.blob is not None and live_req.blob.mime_type == "image/jpeg":
                last_valid_req = live_req

        # 如果找到有效图片，则处理
        if last_valid_req is not None:
            print("Processing the most recent frame from the queue")

            # 用 blob 的数据和 mime 类型创建图片 part
            image_part = genai_types.Part.from_bytes(
                data=last_valid_req.blob.data, mime_type=last_valid_req.blob.mime_type
            )

            contents = genai_types.Content(
                role="user",
                parts=[image_part, genai_types.Part.from_text(prompt_text)],
            )

            # 调用模型根据图片和提示生成内容
            response = client.models.generate_content(
                model="gemini-2.0-flash-exp",
                contents=contents,
                config=genai_types.GenerateContentConfig(
                    system_instruction=(
                        "You are a helpful video analysis assistant. You can count"
                        " the number of people in this image or video. Just respond"
                        " with a numeric number."
                    )
                ),
            )
            if not last_count:
                last_count = response.candidates[0].content.parts[0].text
            elif last_count != response.candidates[0].content.parts[0].text:
                last_count = response.candidates[0].content.parts[0].text
                yield response
                print("response:", response)

        # 等待后再检查新图片
        await asyncio.sleep(0.5)


# 使用此函数帮助 ADK 在需要时停止你的流式工具。
# 例如，如果我们想停止 `monitor_stock_price`，智能体会
# 调用此函数 stop_streaming(function_name=monitor_stock_price)。
def stop_streaming(function_name: str):
    """停止流式

    参数：
      function_name: 要停止的流式函数名。
    """
    pass


root_agent = Agent(
    model="gemini-2.0-flash-exp",
    name="video_streaming_agent",
    instruction="""
      你是一个监控智能体。你可以使用提供的工具/函数进行视频监控和股票价格监控。
      当用户想要监控视频流时，
      你可以使用 monitor_video_stream 函数。当 monitor_video_stream
      返回警报时，你应该告知用户。
      当用户想要监控股票价格时，你可以使用 monitor_stock_price。
      不要问太多问题。不要太啰嗦。
    """,
    tools=[
        monitor_video_stream,
        monitor_stock_price,
        FunctionTool(stop_streaming),
    ],
)



from achatbot.services.a2a_multiagents import ADKBaseHostAgent


class ADKPlanerAgent(ADKBaseHostAgent):
    f"""
    https://github.com/weedge/doraemon-nb/blob/main/langchain/plan_and_execute.ipynb
    !TODO
    """



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from inspect import isfunction
from operator import mul
from functools import reduce, wraps

from aml.multimodal_video.utils.einops.lib import rearrange, repeat
from aml.multimodal_video.utils.einops.lib.layers.torch import Rearrange

from fairseq.modules.local_attention import LocalAttention

# constants

TOKEN_SELF_ATTN_VALUE = -5e4
KMEAN_INIT_ITERS = 10

# helper functions


def exists(val):
    return val is not None


def identity(x, *args, **kwargs):
    return x


def default(x, d):
    if not exists(x):
        return d if not isfunction(d) else d()
    return x


def cast_tuple(x):
    return x if isinstance(x, tuple) else (x,)


def cache_fn(f):
    cache = None

    @wraps(f)
    def cached_fn(*args, **kwargs):
        nonlocal cache
        if exists(cache):
            return cache
        cache = f(*args, **kwargs)
        return cache
    return cached_fn


def to(t):
    return {'device': t.device, 'dtype': t.dtype}


def find_modules(nn_module, type):
    return [module for module in nn_module.modules() if isinstance(module, type)]


def is_empty(t):
    return t.nelement() == 0


def max_neg_value(tensor):
    return -torch.finfo(tensor.dtype).max


def batched_index_select(values, indices):
    last_dim = values.shape[-1]
    return values.gather(2, expand_dim(indices, -1, last_dim))


def merge_dims(ind_from, ind_to, tensor):
    shape = list(tensor.shape)
    arr_slice = slice(ind_from, ind_to + 1)
    shape[arr_slice] = [reduce(mul, shape[arr_slice])]
    return tensor.reshape(*shape)


def expand_dim(t, dim, k):
    t = t.unsqueeze(dim)
    expand_shape = [-1] * len(t.shape)
    expand_shape[dim] = k
    return t.expand(*expand_shape)


def scatter_mean(src, t, index, dim, eps=1e-5):
    numer = src.scatter_add(dim, index, t)
    denom = src.scatter_add(dim, index, torch.ones_like(t))
    return numer / (denom + eps)


def split_at_index(dim, index, t):
    pre_slices = (slice(None),) * dim
    l = (*pre_slices, slice(None, index))
    r = (*pre_slices, slice(index, None))
    return t[l], t[r]


def reshape_dim(t, dim, split_dims):
    shape = list(t.shape)
    num_dims = len(shape)
    dim = (dim + num_dims) % num_dims
    shape[dim:dim+1] = split_dims
    return t.reshape(shape)


def ema(old, new, decay):
    if not exists(old):
        return new
    return old * decay + new * (1 - decay)


def ema_inplace(moving_avg, new, decay):
    if is_empty(moving_avg):
        moving_avg.data.copy_(new)
        return
    moving_avg.data.mul_(decay).add_(new, alpha=(1 - decay))

# helper classes


def map_first_tuple_or_el(x, fn):
    if isinstance(x, tuple):
        return (fn(x[0]),) + x[1:]
    return fn(x)


class Chunk(nn.Module):
    def __init__(self, chunks, fn, along_dim=-1):
        super().__init__()
        self.dim = along_dim
        self.chunks = chunks
        self.fn = fn

    def forward(self, x, **kwargs):
        if self.chunks <= 1:
            return self.fn(x, **kwargs)
        chunks = x.chunk(self.chunks, dim=self.dim)
        return torch.cat([self.fn(c, **kwargs) for c in chunks], dim=self.dim)


class PreNorm(nn.ModuleList):
    def __init__(self, norm_class, dim, fn):
        super().__init__()
        self.norm = norm_class(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        x = self.norm(x)
        return self.fn(x, **kwargs)


class ReZero(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.residual_weight = nn.Parameter(torch.zeros(1))
        self.fn = fn

    def forward(self, x, **kwargs):
        x = self.fn(x, **kwargs)
        return map_first_tuple_or_el(x, lambda t: t * self.residual_weight)


class ScaleNorm(nn.Module):
    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.g = nn.Parameter(torch.ones(1))
        self.eps = eps

    def forward(self, x):
        def norm(t):
            n = torch.norm(t, dim=-1, keepdim=True).clamp(min=self.eps)
            return t / n * self.g
        return map_first_tuple_or_el(x, norm)


class ProjectInOut(nn.Module):
    def __init__(self, fn, dim_in, dim_out, project_out=True):
        super().__init__()
        self.fn = fn
        self.project_in = nn.Linear(dim_in, dim_out)
        self.project_out = nn.Linear(dim_out, dim_in) if project_out else identity

    def forward(self, x, **kwargs):
        x = self.project_in(x)
        x, loss = self.fn(x, **kwargs)
        x = self.project_out(x)
        return x, loss


class MatrixMultiply(nn.Module):
    def __init__(self, tensor, transpose=False):
        super().__init__()
        self.tensor = tensor
        self.transpose = transpose

    def forward(self, x):
        tensor = self.tensor
        if self.transpose:
            tensor = tensor.t()
        return x @ tensor

# positional embeddings


class DepthWiseConv1d(nn.Module):
    def __init__(self, dim_in, dim_out, kernel_size, stride=1, bias=True, causal=False):
        super().__init__()
        self.padding = ((kernel_size - 1), 0) if causal else (kernel_size // 2, kernel_size // 2)

        self.net = nn.Sequential(
            nn.Conv1d(dim_in, dim_in, kernel_size=kernel_size, groups=dim_in, stride=stride, bias=bias),
            nn.Conv1d(dim_in, dim_out, 1, bias=bias)
        )

    def forward(self, x):
        x = F.pad(x, self.padding, value=0.)
        return self.net(x)


class FixedPositionalEmbedding(nn.Module):
    def __init__(self, dim, max_seq_len):
        super().__init__()
        inv_freq = 1. / (10000 ** (torch.arange(0, dim, 2).float() / dim))
        position = torch.arange(0, max_seq_len, dtype=torch.float)
        sinusoid_inp = torch.einsum("i,j->ij", position, inv_freq)
        emb = torch.cat((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)
        self.register_buffer('emb', emb)

    def forward(self, x):
        return self.emb[None, :x.shape[1], :].to(x)


def rotate_every_two(x):
    x = rearrange(x, '... (d j) -> ... d j', j=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, '... d j -> ... (d j)')


def apply_rotary_pos_emb(q, k, sinu_pos):
    sinu_pos = rearrange(sinu_pos, '() n (j d) -> n j d', j=2)
    sin, cos = sinu_pos.unbind(dim=-2)
    sin, cos = map(lambda t: repeat(t, 'b n -> b (n j)', j=2), (sin, cos))
    q, k = map(lambda t: (t * cos) + (rotate_every_two(t) * sin), (q, k))
    return q, k

# kmeans related function and class


def update_kmeans_on_backwards(module):
    module.kmean_modules = find_modules(module, Kmeans)

    def hook(_, grad_in, grad_out):
        for m in module.kmean_modules:
            m.update()

    return module.register_backward_hook(hook)


def similarity(x, means):
    return torch.einsum('bhld,hcd->bhlc', x, means)


def dists_and_buckets(x, means):
    dists = similarity(x, means)
    _, buckets = torch.max(dists, dim=-1)
    return dists, buckets


def batched_bincount(index, num_classes, dim=-1):
    shape = list(index.shape)
    shape[dim] = num_classes
    out = index.new_zeros(shape)
    out.scatter_add_(dim, index, torch.ones_like(index, dtype=index.dtype))
    return out


def kmeans_iter(x, means, buckets=None):
    b, h, _, d, dtype, num_clusters = *x.shape, x.dtype, means.shape[1]

    if not exists(buckets):
        _, buckets = dists_and_buckets(x, means)

    bins = batched_bincount(buckets, num_clusters).sum(0, keepdim=True)
    zero_mask = bins.long() == 0

    means_ = buckets.new_zeros(b, h, num_clusters, d, dtype=dtype)
    means_.scatter_add_(-2, expand_dim(buckets, -1, d), x)
    means_ = F.normalize(means_.sum(0, keepdim=True), dim=-1).type(dtype)

    means = torch.where(zero_mask.unsqueeze(-1), means, means_)
    means = means.squeeze(0)
    return means


def distribution(dists, window_size):
    _, topk_indices = dists.topk(k=window_size, dim=-2)
    indices = topk_indices.transpose(-2, -1)
    return indices.reshape(*indices.size()[:2], -1)


class Kmeans(nn.Module):
    def __init__(self, num_heads, head_dim, num_clusters, ema_decay=0.999, commitment=1e-4):
        super().__init__()
        self.commitment = commitment
        self.ema_decay = ema_decay

        self.register_buffer('means', torch.randn(num_heads, num_clusters, head_dim))
        self.register_buffer('initted', torch.tensor(False))
        self.num_new_means = 0
        self.new_means = None

    @torch.no_grad()
    def init(self, x):
        if self.initted:
            return
        _, h, _, d, device, _ = *x.shape, x.device, x.dtype

        num_clusters = self.means.shape[1]

        means = x.transpose(0, 1).contiguous().view(h, -1, d)
        num_samples = means.shape[1]

        if num_samples >= num_clusters:
            indices = torch.randperm(num_samples, device=device)[:num_clusters]
        else:
            indices = torch.randint(0, num_samples, (num_clusters,), device=device)

        means = means[:, indices]

        for _ in range(KMEAN_INIT_ITERS):
            means = kmeans_iter(x, means)

        self.num_new_means = 0
        self.means.data.copy_(means)
        self.initted.data.copy_(torch.tensor(True))

    @torch.no_grad()
    def update(self, new_means=None):
        new_means = default(new_means, self.new_means)
        assert exists(new_means), 'new kmeans has not been supplied'
        ema_inplace(self.means, new_means, self.ema_decay)

        del self.new_means
        self.new_means = None
        self.num_new_means = 0

    def forward(self, x, update_means=False):
        self.init(x)

        b, dtype = x.shape[0], x.dtype
        means = self.means.type(dtype)
        x = F.normalize(x, 2, dim=-1).type(dtype)

        with torch.no_grad():
            dists, buckets = dists_and_buckets(x, means)

        routed_means = batched_index_select(expand_dim(means, 0, b), buckets)
        loss = F.mse_loss(x, routed_means) * self.commitment

        if update_means:
            with torch.no_grad():
                means = kmeans_iter(x, means, buckets)
            self.new_means = ema(self.new_means, means, self.num_new_means / (self.num_new_means + 1))
            self.num_new_means += 1

        return dists, loss

# kmeans attention class


class KmeansAttention(nn.Module):
    def __init__(self, num_clusters, window_size, num_heads, head_dim, causal=False, dropout=0., ema_decay=0.999, commitment=1e-4, context_window_size=None, receives_context=False, num_mem_kv=0, shared_qk=False):
        super().__init__()
        self.num_heads = num_heads
        self.num_clusters = num_clusters
        self.head_dim = head_dim

        self.window_size = window_size
        self.context_window_size = default(context_window_size, window_size)
        self.causal = causal

        self.shared_qk = shared_qk
        self.receives_context = receives_context
        self.kmeans = Kmeans(num_heads, head_dim, num_clusters, ema_decay, commitment)
        self.dropout = nn.Dropout(dropout)

        self.num_mem_kv = max(num_mem_kv, 1 if causal and not shared_qk else 0)
        self.mem_key = nn.Parameter(torch.randn(num_heads, num_clusters, self.num_mem_kv, head_dim))
        self.mem_value = nn.Parameter(torch.randn(num_heads, num_clusters, self.num_mem_kv, head_dim))

    def forward(self, q, k, v, query_mask=None, key_mask=None, **kwargs):
        b, h, t, d, kv_t, wsz, c_wsz, nc, device, dtype = *q.shape, k.shape[2], self.window_size, self.context_window_size, self.num_clusters, q.device, q.dtype
        is_reverse = kwargs.pop('_reverse', False)

        out = torch.zeros_like(q, dtype=dtype)

        update_kmeans = self.training and not is_reverse

        key_mask = default(key_mask, query_mask) if not self.receives_context else key_mask
        kv_wsz = wsz if not self.receives_context else c_wsz

        wsz = min(wsz, t)
        kv_wsz = min(kv_wsz, kv_t)

        if not self.shared_qk or self.receives_context:
            dists, aux_loss = self.kmeans(torch.cat((q, k), dim=2), update_kmeans)
            q_dists, k_dists = split_at_index(2, t, dists)
            indices = distribution(q_dists, wsz)
            kv_indices = distribution(k_dists, kv_wsz)
        else:
            dists, aux_loss = self.kmeans(q, update_kmeans)
            k = F.normalize(k, dim=-1).to(q)
            indices = distribution(dists, wsz)
            kv_indices = indices

        q = batched_index_select(q, indices)
        k = batched_index_select(k, kv_indices)
        v = batched_index_select(v, kv_indices)

        reshape_with_window = lambda x: x.reshape(b, h, nc, -1, d)
        q, k, v = map(reshape_with_window, (q, k, v))

        m_k, m_v = map(lambda x: expand_dim(x, 0, b).to(q), (self.mem_key, self.mem_value))
        k, v = map(lambda x: torch.cat(x, dim=3), ((m_k, k), (m_v, v)))

        dots = torch.einsum('bhnid,bhnjd->bhnij', q, k) * (d ** -0.5)

        mask_value = max_neg_value(dots)

        if exists(query_mask) or exists(key_mask):
            query_mask = default(query_mask, lambda: torch.ones((b, t), device=device).bool())
            key_mask = default(key_mask, lambda: torch.ones((b, kv_t), device=device).bool())

            q_mask = expand_dim(query_mask, 1, h).gather(2, indices)
            kv_mask = expand_dim(key_mask, 1, h).gather(2, kv_indices)
            q_mask, kv_mask = map(lambda t: t.reshape(b, h, nc, -1), (q_mask, kv_mask))
            mask = q_mask[:, :, :, :, None] * kv_mask[:, :, :, None, :]
            mask = F.pad(mask, (self.num_mem_kv, 0), value=1)
            dots.masked_fill_(~mask, mask_value)
            del mask

        if self.causal:
            q_mask, kv_mask = map(lambda t: t.reshape(b, h, nc, -1), (indices, kv_indices))
            mask = q_mask[:, :, :, :, None] >= kv_mask[:, :, :, None, :]
            mask = F.pad(mask, (self.num_mem_kv, 0), value=1)
            dots.masked_fill_(~mask, mask_value)
            del mask

        if self.shared_qk:
            q_mask, kv_mask = map(lambda t: t.reshape(b, h, nc, -1), (indices, kv_indices))
            mask = q_mask[:, :, :, :, None] == kv_mask[:, :, :, None, :]
            mask = F.pad(mask, (self.num_mem_kv, 0), value=0)
            dots.masked_fill_(mask, TOKEN_SELF_ATTN_VALUE)
            del mask

        dots = dots.softmax(dim=-1)
        dots = self.dropout(dots)

        bo = torch.einsum('bhcij,bhcjd->bhcid', dots, v)
        so = torch.reshape(bo, (b, h, -1, bo.shape[-1])).type(dtype)
        out = scatter_mean(out, so, indices.unsqueeze(-1).expand_as(so), -2)
        return out, aux_loss

# feedforward


class GELU_(nn.Module):
    def forward(self, x):
        return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))


GELU = nn.GELU if hasattr(nn, 'GELU') else GELU_


class FeedForward(nn.Module):
    def __init__(self, dim, mult=4, dropout=0., activation=None, glu=False):
        super().__init__()
        activation = default(activation, GELU)

        self.glu = glu
        self.w1 = nn.Linear(dim, dim * mult * (2 if glu else 1))
        self.act = activation()
        self.dropout = nn.Dropout(dropout)
        self.w2 = nn.Linear(dim * mult, dim)

    def forward(self, x, **kwargs):
        if not self.glu:
            x = self.w1(x)
            x = self.act(x)
        else:
            x, v = self.w1(x).chunk(2, dim=-1)
            x = self.act(x) * v

        x = self.dropout(x)
        x = self.w2(x)
        return x

# self attention


class SelfAttention(nn.Module):
    def __init__(self, dim, max_seq_len, heads, local_attn_heads, window_size, dim_head=None, local_attn_window_size=None, local_attn_radius_blocks=1, causal=False, attn_dropout=0., dropout=0., kmeans_ema_decay=0.999, commitment_factor=1e-4, receives_context=False, context_window_size=None, rel_pos_emb=True, num_mem_kv=0, shared_qk=False, conv_query_kernel=9):
        super().__init__()
        assert dim_head or (dim % heads) == 0, 'hidden dimension must be divisible by number of heads'
        assert (max_seq_len % window_size) == 0, 'maximum sequence length must be divisible by the target window size'
        assert local_attn_heads <= heads, 'number of local attention heads must be less than total heads'
        assert not (receives_context and local_attn_heads > 0), 'local attention cannot be used for self attention with context'
        assert not (receives_context and causal), 'contextual attention layer cannot be causal'

        local_attn_window_size = default(local_attn_window_size, window_size)
        context_window_size = default(context_window_size, window_size)

        self.shared_qk = shared_qk
        self.receives_context = receives_context
        self.heads = heads
        self.local_attn_heads = local_attn_heads
        self.global_attn_heads = heads - local_attn_heads

        self.causal = causal
        self.window_size = window_size

        dim_head = default(dim_head, dim // heads)
        dim_heads = dim_head * heads
        self.dim_head = dim_head

        num_clusters = max_seq_len // window_size

        # local

        local_dim_heads = dim_head * self.local_attn_heads

        if self.local_attn_heads > 0:
            rel_pos_emb_config = (dim_head, local_attn_heads) if rel_pos_emb else None
            self.local_attn = LocalAttention(dim=dim_head, window_size=local_attn_window_size, causal=causal, dropout=attn_dropout, rel_pos_emb_config=rel_pos_emb_config, look_backward=local_attn_radius_blocks, look_forward=0 if causal else local_attn_radius_blocks)
            self.local_to_qkv = nn.Linear(dim, 3 * local_dim_heads)

        # global

        global_dim_heads = dim_head * self.global_attn_heads

        if self.global_attn_heads > 0:
            self.global_attn = KmeansAttention(num_clusters, window_size, self.global_attn_heads, dim_head, causal=causal, dropout=attn_dropout, ema_decay=kmeans_ema_decay, commitment=commitment_factor, receives_context=receives_context, num_mem_kv=num_mem_kv, shared_qk=shared_qk)

        self.to_q = nn.Sequential(
            Rearrange('b n c -> b c n'),
            DepthWiseConv1d(dim, global_dim_heads, conv_query_kernel, causal=causal),
            Rearrange('b c n -> b n c')
        )

        self.to_v = nn.Linear(dim, global_dim_heads, bias=False)

        if not self.shared_qk:
            self.to_k = nn.Linear(dim, global_dim_heads, bias=False)

        # out

        self.to_out = nn.Linear(dim_heads, dim, bias=False)
        self.dropout = nn.Dropout(dropout)

    def forward(self, query, key, value, context=None, key_padding_mask=None, context_mask=None, pos_emb=None, **kwargs):
        assert not (self.receives_context and not exists(context)), 'context must be passed if self attention is set to receive context'
        input_mask = key_padding_mask
        x = query.transpose(0, 1)
        b, t, _, h, dh = *x.shape, self.heads, self.dim_head
        has_local, has_global = map(lambda x: x > 0, (self.local_attn_heads, self.global_attn_heads))

        split_heads = lambda v: reshape_dim(v, -1, (-1, dh)).transpose(1, 2).contiguous()

        if has_local:
            local_qkv = self.local_to_qkv(x).chunk(3, dim=-1)
            lq, lk, lv = map(split_heads, local_qkv)

        if has_global:
            kv_input = x if not self.receives_context else context

            q, v = self.to_q(x), self.to_v(kv_input)

            if not self.shared_qk:
                k = self.to_k(kv_input)
            else:
                k = self.to_q(kv_input) if self.receives_context else q

            q, k, v = map(split_heads, (q, k, v))

        out = []
        total_loss = torch.tensor(0., requires_grad=True, **to(x))

        if has_local:
            local_out = self.local_attn(lq, lk, lv, input_mask=input_mask)
            out.append(local_out)

        if has_global:
            if not self.receives_context and exists(pos_emb):
                q, k = apply_rotary_pos_emb(q, k, pos_emb)

            global_out, loss = self.global_attn(q, k, v, query_mask=input_mask, key_mask=context_mask)
            total_loss = total_loss + loss

            out.append(global_out)

        out = torch.cat(out, dim=1)
        out = out.reshape(b, h, t, -1).transpose(1, 2).reshape(b, t, -1)
        out = self.dropout(out.transpose(0, 1))
        # out = self.to_out(out)
        return out, total_loss



import os
import sys
import logging
from threading import Thread
from typing import BinaryIO

try:
    import torch
    import numpy

    from transformers import GenerationConfig

    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../../../StepAudio2"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../../../deps/StepAudio2"))

    from achatbot.StepAudio2.stepaudio2 import StepAudio2, StepAudio2Base
    from achatbot.StepAudio2.utils import (
        compute_token_num,
        load_audio,
        log_mel_spectrogram,
        padding_mels,
    )

    from achatbot.core.llm.transformers.streamer import TokenStreamer
    from achatbot.common.utils.helper import get_device

except ModuleNotFoundError as e:
    raise Exception(f"Missing module: {e}")


class StepAudio2StreamBase(StepAudio2Base):
    def __init__(self, model_path: str, verbose: bool = False):
        super().__init__(model_path)
        self._verbose = verbose

    def apply_chat_template(self, messages: list):
        """
        add np.ndarray/torch.Tensor audio msg support
        - audio sample rate: 16000
        """
        results = []
        mels = []
        for msg in messages:
            content = msg
            if isinstance(content, str):
                text_with_audio = content
                results.append(text_with_audio)
            elif isinstance(content, dict):
                if content["type"] == "text":
                    results.append(f"{content['text']}")
                elif content["type"] == "audio":
                    audio = content["audio"]
                    if isinstance(audio, (BinaryIO, str, os.PathLike)):
                        audio = load_audio(audio)
                    elif isinstance(audio, numpy.ndarray):
                        audio = torch.from_numpy(audio, dtype=torch.float32)
                    assert isinstance(audio, torch.Tensor), f"Unsupported audio type: {type(audio)}"
                    if len(audio.shape) > 1:  # [1, size]
                        audio = audio.squeeze(0)  # [size]
                    for i in range(0, audio.shape[0], 16000 * 25):
                        mel = log_mel_spectrogram(
                            audio[i : i + 16000 * 25], n_mels=128, padding=479
                        )
                        mels.append(mel)
                        audio_tokens = "<audio_patch>" * compute_token_num(mel.shape[1])
                        results.append(f"<audio_start>{audio_tokens}<audio_end>")
                elif content["type"] == "token":
                    results.append(content["token"])
            else:
                raise ValueError(f"Unsupported content type: {type(content)}")
        return results, mels

    def __call__(self, messages: list, **kwargs):
        messages, mels = self.apply_chat_template(messages)
        if self._verbose:
            print(f"messages: {messages}")
            if len(mels) > 0:
                print(f"{len(mels)=} {mels[0].shape=}")

        # Tokenize prompts
        prompt_ids = []
        for msg in messages:
            if isinstance(msg, str):
                prompt_ids.append(
                    self.llm_tokenizer(text=msg, return_tensors="pt", padding=True)["input_ids"]
                )
            elif isinstance(msg, list):
                prompt_ids.append(torch.tensor([msg], dtype=torch.int32))
            else:
                raise ValueError(f"Unsupported content type: {type(msg)}")
        prompt_ids = torch.cat(prompt_ids, dim=-1).cuda()
        attention_mask = torch.ones_like(prompt_ids)

        # mels = None if len(mels) == 0 else torch.stack(mels).cuda()
        # mel_lengths = None if mels is None else torch.tensor([mel.shape[1] - 2 for mel in mels], dtype=torch.int32, device='cuda')
        if len(mels) == 0:
            mels = None
            mel_lengths = None
        else:
            mels, mel_lengths = padding_mels(mels)
            mels = mels.cuda()
            mel_lengths = mel_lengths.cuda()

        generation_config = dict(
            # max_new_tokens=256,
            pad_token_id=self.llm_tokenizer.pad_token_id,
            eos_token_id=self.eos_token_id,
        )
        generation_config.update(kwargs)
        generation_config = GenerationConfig(**generation_config)
        logging.debug(f"generation_config: {generation_config}")

        streamer = TokenStreamer(skip_prompt=True)

        generation_kwargs = dict(
            input_ids=prompt_ids,
            wavs=mels,
            wav_lens=mel_lengths,
            attention_mask=attention_mask,
            generation_config=generation_config,
            streamer=streamer,
            tokenizer=self.llm_tokenizer,
        )

        thread = Thread(target=self.llm.generate, kwargs=generation_kwargs)
        thread.start()

        stop_ids = (
            [generation_config.eos_token_id]
            if isinstance(generation_config.eos_token_id, int)
            else generation_config.eos_token_id
        )
        for token_id in streamer:
            if token_id in stop_ids:
                break
            yield token_id


class StepAudio2Stream(StepAudio2StreamBase):
    def apply_chat_template(self, messages: list):
        """
        add np.ndarray/torch.Tensor audio msg support
        - audio sample rate: 16000
        """
        results = []
        mels = []
        for msg in messages:
            role = msg["role"]
            content = msg["content"]
            if role == "user":
                role = "human"
            if isinstance(content, str):
                text_with_audio = f"<|BOT|>{role}\n{content}"
                text_with_audio += "<|EOT|>" if msg.get("eot", True) else ""
                results.append(text_with_audio)
            elif isinstance(content, list):
                results.append(f"<|BOT|>{role}\n")
                for item in content:
                    if item["type"] == "text":
                        results.append(f"{item['text']}")
                    elif item["type"] == "audio":
                        audio = item["audio"]
                        if isinstance(audio, (BinaryIO, str, os.PathLike)):
                            audio = load_audio(audio)
                        elif isinstance(audio, numpy.ndarray):
                            audio = torch.from_numpy(audio, dtype=torch.float32)
                        assert isinstance(audio, torch.Tensor), (
                            f"Unsupported audio type: {type(audio)}"
                        )
                        if len(audio.shape) > 1:  # [1, size]
                            audio = audio.squeeze(0)  # [size]
                        for i in range(0, audio.shape[0], 16000 * 25):
                            mel = log_mel_spectrogram(
                                audio[i : i + 16000 * 25], n_mels=128, padding=479
                            )
                            mels.append(mel)
                            audio_tokens = "<audio_patch>" * compute_token_num(mel.shape[1])
                            results.append(f"<audio_start>{audio_tokens}<audio_end>")
                    elif item["type"] == "token":
                        results.append(item["token"])
                if msg.get("eot", True):
                    results.append("<|EOT|>")
            elif content is None:
                results.append(f"<|BOT|>{role}\n")
            else:
                raise ValueError(f"Unsupported content type: {type(content)}")
        # print(results)
        return results, mels



import glob
import os
import sys
import time
import json
import queue
import asyncio
import logging
import threading
from typing import AsyncGenerator, List

import uuid
import torch
from apipeline.frames import *

try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../StepAudio2"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../deps/StepAudio2"))

    from achatbot.StepAudio2.token2wav import Token2wav
except ModuleNotFoundError as e:
    raise Exception(f"Missing module: {e}")

from achatbot.common.utils.time import time_now_iso8601
from achatbot.common.interface import ILlm
from achatbot.common.types import ASSETS_DIR, RECORDS_DIR
from achatbot.processors.voice.base import VoiceProcessorBase
from achatbot.common.session import Session
from achatbot.common.types import SessionCtx
from achatbot.common.utils.audio_utils import bytes2TorchTensorWith16
from achatbot.types.frames import (
    Language,
    TextQuestionsAudioRawFrame,
    PathAudioRawFrame,
    LLMGenedTokensFrame,
    FunctionCallFrame,
    ReasoningThinkTextFrame,
    TranscriptionFrame,
    VADStateAudioRawFrame,
)
import achatbot.modules.functions.search.api
import achatbot.modules.functions.weather.api
from achatbot.modules.functions.function import FunctionManager
from achatbot.core.llm.transformers.manual_voice_step2 import TransformersManualVoiceStep2
from achatbot.core.llm.vllm.step_audio2 import VllmClientStepAudio2
from .helper import extract_function_info


class StepAudio2BaseProcessor(VoiceProcessorBase):
    """ """

    CHUNK_SIZE = 25
    SYS_PROMPT = "You are a helpful assistant."

    def __init__(
        self,
        *,
        init_system_prompt: str = "",
        text_stream_out: bool = False,
        prompt_wav: str = "",
        warmup_cn: int = 1,
        chat_history_size: int | None = None,
        no_stream_sleep_time: float = 0.5,
        chunk_size: int = 0,
        session: Session | None = None,
        audio_llm: ILlm | None = None,
        token2wav: Token2wav | None = None,
        is_speaking: bool = True,
        tools: list = [],
        verbose: bool = False,
        set_last_chunk: bool = True,
        is_reasoning_think: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        assert audio_llm is not None, "audio_llm is None"

        assert isinstance(audio_llm, (VllmClientStepAudio2, TransformersManualVoiceStep2)), (
            "audio_llm is not TransformersManualVoiceStep2 or VllmClientStepAudio2"
        )

        self._audio_llm = audio_llm
        self._is_speaking = is_speaking
        if is_speaking is True:
            token2wav_path = os.path.join(audio_llm.args.lm_model_name_or_path, "token2wav")
            self._prompt_wav = prompt_wav or os.path.join(ASSETS_DIR, "default_female.wav")
            self._token2wav = token2wav or Token2wav(
                token2wav_path,
                prompt_wav=self._prompt_wav,
                warmup_cn=warmup_cn,
                verbose=verbose,
                **kwargs,
            )

        self._system_prompt = init_system_prompt or self.SYS_PROMPT
        self._text_stream_out = text_stream_out
        self._chunk_size = chunk_size or self.CHUNK_SIZE

        self._session = session or Session(**SessionCtx(str(uuid.uuid4())).__dict__)
        self._session.set_chat_history_size(chat_history_size)
        self._session.chat_history.init({"role": "system", "content": self._system_prompt})
        self._tools = FunctionManager.get_tool_calls_by_names(tools)
        if len(self._tools) > 0 and isinstance(self._audio_llm, TransformersManualVoiceStep2):
            tool_json_schemas = json.dumps(self._tools)
            self._session.chat_history.init_tools(
                {"role": "tool_json_schemas", "content": tool_json_schemas}
            )

        # for async generate to yield
        self._queue = queue.Queue()
        self._input_queue = queue.Queue()
        self._generate_thread = None
        self._generate_over = False

        self._sleep_time = no_stream_sleep_time
        self._verbose = verbose
        self._set_last_chunk = set_last_chunk
        # self._is_push_frame = kwargs.get("is_push_frame", False)
        self._is_push_frame = False  # close this param, just debug

        # think
        self._is_reasoning_think = is_reasoning_think

    @property
    def chat_history(self):
        return self._session.chat_history

    def reset(self):
        if self._is_speaking is True:
            self._token2wav.cache = {}
        self._session.reset()

    def _generate(self):
        while True:
            try:
                item = self._input_queue.get()
                if item is None:
                    self._queue.put(None)  # Signal the end of the stream
                    break  # Signal to stop the thread
                session, kwargs = item
                if self._tools:
                    kwargs["tools"] = self._tools

                if kwargs.get("stop") and "</think>" in kwargs.get("stop"):
                    is_think = True
                elif kwargs.get("stop_strings") and "</think>" in kwargs.get("stop_strings"):
                    is_think = True
                else:
                    is_think = False

                if self._verbose:
                    print(f"{session=} {kwargs=} {is_think=}")

                token_iter = self._audio_llm.generate(session, **kwargs)
                _, _, is_tool_call = self.put_out_audio_text(
                    token_iter, is_out_text=True, is_think=is_think
                )
                if (
                    is_tool_call is False and is_think is False
                ):  # no tool call, no think end round generate
                    self._queue.put(None)  # Signal the end of the stream

                # reset token2wav stream cache
                if self._is_speaking is True and is_think is False:
                    start = time.time()
                    self._token2wav.set_stream_cache(self._prompt_wav)
                    print(f"token2wav.set_stream_cache cost: {time.time() - start}")
            except Exception as e:
                logging.error(f"Exception generate: {e}", exc_info=True)
                self._queue.put(None)  # Signal the end of the stream
                break

    async def start(self, frame: StartFrame):
        await super().start(frame)
        self._generate_thread = threading.Thread(target=self._generate)
        self._generate_thread.daemon = True
        self._generate_thread.start()
        logging.info("start done")

    async def stop(self, frame: EndFrame):
        await super().stop(frame)
        self._input_queue.put(None)  # Signal the thread to stop
        self._generate_over = True
        self._generate_thread.join()  # Wait for the thread to finish
        logging.info("stop done")

    async def cancel(self, frame: CancelFrame):
        await super().cancel(frame)
        self._input_queue.put(None)  # Signal the thread to stop
        self._generate_over = True
        self._generate_thread.join()  # Wait for the thread to finish
        logging.info("cancel done")

    async def gen(self, is_push_frame: bool = False) -> AsyncGenerator[Frame, None]:
        while True:
            try:
                item = self._queue.get_nowait()
                if item is None:
                    logging.info(f"generate done")
                    break  # End of the stream
                # logging.info(f"generate data: {item}")
                if is_push_frame is True or self._is_push_frame is True:
                    await self.push_frame(item)
                    yield None
                else:
                    yield item
            except queue.Empty:
                # yield asysncio.sleep to allow other tasks to run, e.g.: sink task (write audio)
                await asyncio.sleep(self._sleep_time)
                # logging.info(f"queue empty sleep {self._sleep_time}")
                continue

    def send_input(self, session: Session, **kwargs):
        self._input_queue.put((session, kwargs))

    @property
    def stream_info(self) -> dict:
        """Return dict out stream info"""
        return {
            "sample_rate": self._audio_llm.RATE,
            "channels": 1,
        }

    async def say(
        self,
        text: str,
        system_prompt: str = "以自然的语速读出下面的文字。\n",
        **kwargs,
    ):
        async for item in self.generator_say(
            text,
            system_prompt=system_prompt,
            is_push_frame=True,
            **kwargs,
        ):
            pass

    async def generator_say(
        self,
        text: str,
        system_prompt: str = "以自然的语速读出下面的文字。\n",
        is_push_frame: str = True,
        **kwargs,
    ) -> AsyncGenerator[Frame, None]:
        """
        support: en,zh,ja
        """
        logging.info(f"say: {text}")

        messages = [
            {"role": "system", "content": system_prompt or self._system_prompt},
            {"role": "human", "content": text},
            {
                "role": "assistant",
                "content": "<tts_start>",
                "eot": False,
            },  # Insert <tts_start> for speech response
        ]
        self._session.ctx.state["messages"] = messages
        self.send_input(self._session, **kwargs)
        async for item in self.gen(is_push_frame=is_push_frame):
            yield item

    def put_out_audio_text(self, token_iter, is_out_text: bool = True, is_think: bool = False):
        output_token_ids = []
        output_audio_token_ids = []
        out_text_token_ids = []
        is_tool = False
        tool_calls_token_ids = []
        is_tag = False
        buffer = []
        unicode_token_id = []
        tool_calls: List[FunctionCallFrame] = []  # for v1/chat/completions api

        for token_id in token_iter:
            if self._generate_over is True:
                break
            if isinstance(token_id, dict) and "tool_calls" in token_id:
                for tool_call in token_id["tool_calls"]:
                    tool_calls.append(
                        FunctionCallFrame(
                            tool_call_id=tool_call["id"],
                            type=tool_call["type"],
                            index=tool_call["index"],
                            function_name=tool_call["function"]["name"],
                            arguments=tool_call["function"]["arguments"],
                        )
                    )
                continue
            if self._verbose is True:
                print(f"{token_id=} {self._audio_llm.llm_tokenizer.decode(token_id)=}")
            output_token_ids.append(token_id)
            if token_id < 151688:  # text
                if token_id == 151657:  # <tool_call>
                    is_tool = True
                    continue
                if token_id == 151658:  # </tool_call>
                    is_tool = False
                    continue
                if is_tool:
                    tool_calls_token_ids.append(token_id)
                    continue

                if token_id == 27:  # <
                    is_tag = True
                    continue
                if token_id == 29:  # >
                    is_tag = False
                    continue
                if is_tag:  # <***>
                    continue

                out_text_token_ids.append(token_id)
                if is_out_text is True and self._text_stream_out is True:
                    out_text = self._audio_llm.llm_tokenizer.decode(unicode_token_id + [token_id])
                    if "�" in out_text:
                        unicode_token_id.append(token_id)
                    else:
                        unicode_token_id = []
                        frame = TextFrame(text=out_text)
                        self._queue.put(frame)
            if token_id > 151695 and self._is_speaking is True:  # audio
                audio_token_id = token_id - 151696
                if audio_token_id < 6561:  # remove audio padding
                    output_audio_token_ids.append(token_id)
                    buffer.append(audio_token_id)
                    if len(buffer) >= self._chunk_size + self._token2wav.flow.pre_lookahead_len:
                        out_bytes = self._token2wav.stream(
                            buffer[: self._chunk_size + self._token2wav.flow.pre_lookahead_len],
                            prompt_wav=self._prompt_wav,
                            last_chunk=False,
                        )
                        frame = AudioRawFrame(
                            audio=out_bytes,
                            sample_rate=self._audio_llm.RATE,
                            num_channels=1,
                        )
                        self._queue.put(frame)
                        buffer = buffer[self._chunk_size :]
        if len(buffer) > 0 and self._is_speaking is True:
            logging.info(f"last chunk size: {len(buffer)}")
            out_bytes = self._token2wav.stream(
                buffer, prompt_wav=self._prompt_wav, last_chunk=self._set_last_chunk
            )
            frame = AudioRawFrame(
                audio=out_bytes,
                sample_rate=self._audio_llm.RATE,
                num_channels=1,
            )
            self._queue.put(frame)

        out_text = ""
        if len(out_text_token_ids) > 0 and is_out_text is True:
            out_text = self._audio_llm.llm_tokenizer.decode(out_text_token_ids)
            if "</think>" not in out_text and is_think:  # think mod out text include </think>
                out_text += "</think>"
            if self._is_reasoning_think and is_think:
                self._queue.put(ReasoningThinkTextFrame(text=out_text))
            if self._text_stream_out is False:
                frame = TextFrame(text=out_text)
                self._queue.put(frame)

        out_audio = (
            self._audio_llm.llm_tokenizer.decode(output_audio_token_ids)
            if len(output_audio_token_ids) > 0
            else []
        )

        is_tool_call = False
        if len(tool_calls_token_ids) > 0:
            tool_calls_token = self._audio_llm.llm_tokenizer.decode(tool_calls_token_ids)
            # print(f"{tool_calls_token=}")
            function_name, function_args = extract_function_info(tool_calls_token)
            self._queue.put(FunctionCallFrame(function_name=function_name, arguments=function_args))
            is_tool_call = True
        if len(tool_calls) > 0:
            for tool_call in tool_calls:
                self._queue.put(
                    FunctionCallFrame(
                        function_name=tool_call.function_name, arguments=tool_call.arguments
                    )
                )
            is_tool_call = True

        self._queue.put(
            LLMGenedTokensFrame(
                token_ids=output_token_ids,
                text_tokens=out_text,
                audio_tokens=out_audio,
                tool_calls=tool_calls,
            )
        )

        return output_token_ids, out_text, is_tool_call


# --------------------------------------------------------------------------------


# A1->T2
class StepAudio2TextProcessor(StepAudio2BaseProcessor):
    """
    audio -> audio_LLM -> text
    - A1->T2 (ASR(trancribe), Audio Understanding, S2TT(support: en,zh,ja))
    - system prompt example:
        - ASR: 请记录下你所听到的语音内容。
        - Audio Understanding: Please briefly explain the important events involved in this audio clip.
        - S2TT: 请仔细聆听这段语音，然后将其内容翻译成中文。
    """

    def __init__(self, **kwargs):
        kwargs["is_speaking"] = False
        self.language = kwargs.pop("language", None)
        super().__init__(**kwargs)

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            audio = frame.path
        else:
            audio = bytes2TorchTensorWith16(frame.audio)

        if len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        messages = [
            {"role": "system", "content": self._system_prompt},
            {
                "role": "human",
                "content": [{"type": "audio", "audio": audio}],
            },
            {"role": "assistant", "content": None},
        ]
        self._session.ctx.state["messages"] = messages
        self.send_input(self._session)
        async for item in self.gen():
            if isinstance(frame, VADStateAudioRawFrame) and isinstance(item, TextFrame):
                language = Language(self.language) if self.language else None
                yield TranscriptionFrame(
                    text=item.text,
                    user_id=self._session.ctx.client_id,
                    timestamp=time_now_iso8601(),
                    language=language,
                    speech_id=frame.speech_id,
                    start_at_s=frame.start_at_s,
                    end_at_s=frame.end_at_s,
                )
            else:
                yield item


class StepASRProcessor(StepAudio2TextProcessor):
    SYS_PROMPT = "请记录下你所听到的语音内容。"


class StepAudioCaptionProcessor(StepAudio2TextProcessor):
    SYS_PROMPT = "Please briefly explain the important events involved in this audio clip."


class StepS2TTProcessor(StepAudio2TextProcessor):
    SYS_PROMPT = "请仔细聆听这段语音，然后将其内容翻译成中文。"


# --------------------------------------------------------------------------------


# T1->A2
class StepText2AudioProcessor(StepAudio2BaseProcessor):
    """
    text -> audio_LLM -> audio
    - T1-A2 (TTS)
    """


class StepTTSProcessor(StepText2AudioProcessor):
    async def run_text(self, frame: TextFrame) -> AsyncGenerator[Frame, None]:
        user_input = frame.text.strip()
        async for item in self.generator_say(
            user_input, system_prompt="以自然的语速读出下面的文字。\n"
        ):
            yield item


# --------------------------------------------------------------------------------


# T1-T2A2
class StepText2TextAudioProcessor(StepAudio2BaseProcessor):
    """
    text -> audio_LLM -> audio
    - T1-T2A2 (T2ST)
    """


class StepT2STProcessor(StepText2TextAudioProcessor):
    async def run_text(self, frame: TextFrame) -> AsyncGenerator[Frame, None]:
        user_input = frame.text.strip()
        async for item in self.generator_say(
            user_input,
            system_prompt="请将下面的文本翻译成英文，并用语音播报。\n",
            is_push_frame=False,
        ):
            yield item


# --------------------------------------------------------------------------------


# A1-T2A2
class StepAudio2TextAudioProcessor(StepAudio2BaseProcessor):
    """
    audio -> audio_LLM -> text and audio
    - A1-T2A2 (S2ST, Paralingustic information understanding)
    """

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            audio = frame.path
        else:
            audio = bytes2TorchTensorWith16(frame.audio)

        if len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        messages = [
            {"role": "system", "content": self._system_prompt},
            {
                "role": "human",
                "content": [{"type": "audio", "audio": audio}],
            },
            {
                "role": "assistant",
                "content": "<tts_start>",
                "eot": False,
            },  # Insert <tts_start> for speech response
        ]
        self._session.ctx.state["messages"] = messages
        self.send_input(self._session)
        async for item in self.gen():
            yield item


class StepS2STProcessor(StepAudio2TextAudioProcessor):
    SYS_PROMPT = "请仔细聆听这段语音，然后将其内容翻译成英文并用语音播报。"


class StepParalingusticInformationUnderstandingProcessor(StepAudio2TextAudioProcessor):
    SYS_PROMPT = "请用语音与我交流。"


# --------------------------------------------------------------------------------


# Chat: multi turn TQTA
class StepText2TextChatProcessor(StepAudio2BaseProcessor):
    """
    text -> audio_LLM -> text
    - T1->T2 (Text Query and Text Answer)
    - system prompt example:
        - TQTA: "You are a helpful assistant."
    """

    def __init__(self, **kwargs):
        kwargs["is_speaking"] = False
        super().__init__(**kwargs)

    async def run_text(self, frame: TextFrame) -> AsyncGenerator[Frame, None]:
        user_input = frame.text.strip()

        self._session.chat_history.append(
            {"role": "human", "content": [{"type": "text", "text": user_input}]}
        )
        self._session.chat_history.append({"role": "assistant", "content": None})
        self._session.ctx.state["messages"] = self._session.chat_history.to_list()
        self.send_input(self._session)
        out_text = ""
        async for item in self.gen():
            if isinstance(item, TextFrame):
                out_text += item.text
            yield item
        self._session.chat_history.pop(-1)
        self._session.chat_history.append({"role": "assistant", "content": out_text})
        self._session.increment_chat_round()


# Chat: multi turn AQTA
class StepAudio2TextChatProcessor(StepAudio2BaseProcessor):
    """
    audio -> audio_LLM -> text
    - A1->T2 (Audio Query and Text Answer)
    - system prompt example:
        - AQTA: "You are a helpful assistant."
    """

    def __init__(self, **kwargs):
        kwargs["is_speaking"] = False
        super().__init__(**kwargs)

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            audio = frame.path
        else:
            audio = bytes2TorchTensorWith16(frame.audio)

        if audio is None or len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        self._session.chat_history.append(
            {"role": "human", "content": [{"type": "audio", "audio": audio}]}
        )
        self._session.chat_history.append({"role": "assistant", "content": None})
        self._session.ctx.state["messages"] = self._session.chat_history.to_list()
        self.send_input(self._session)
        out_text = ""
        async for item in self.gen():
            if isinstance(item, TextFrame):
                out_text += item.text
            yield item
        self._session.chat_history.pop(-1)
        self._session.chat_history.append({"role": "assistant", "content": out_text})
        self._session.increment_chat_round()


# Chat: multi turn TQAA
class StepText2TextAudioChatProcessor(StepAudio2BaseProcessor):
    """
    text -> audio_LLM -> audio and text
    - T1-T2A2 (Text Query and Audio Answer)
    - system prompt example:
        - TQAA: "You are a helpful assistant."
    """

    async def run_text(self, frame: TextFrame) -> AsyncGenerator[Frame, None]:
        user_input = frame.text.strip()

        self._session.chat_history.append(
            {"role": "human", "content": [{"type": "text", "text": user_input}]}
        )
        self._session.chat_history.append(
            {
                "role": "assistant",
                "content": "<tts_start>",
                "eot": False,
            }  # Insert <tts_start> for speech response
        )
        self._session.ctx.state["messages"] = self._session.chat_history.to_list()
        self.send_input(self._session)
        async for item in self.gen():
            if isinstance(item, LLMGenedTokensFrame):
                self._session.chat_history.pop(-1)
                if isinstance(self._audio_llm, TransformersManualVoiceStep2):
                    self._session.chat_history.append(
                        {
                            "role": "assistant",
                            "content": [
                                {"type": "text", "text": "<tts_start>"},
                                {"type": "token", "token": item.token_ids},
                            ],
                        }
                    )
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item = {
                        "role": "assistant",
                        "tts_content": {
                            "tts_text": item.text_tokens,
                            "tts_audio": item.audio_tokens,
                        },
                    }
                    if len(item.tool_calls) > 0:
                        history_item["tool_calls"] = []
                        for tool_call in item.tool_calls:
                            history_item["tool_calls"].append(
                                {
                                    "id": tool_call.tool_call_id,
                                    "type": tool_call.type,
                                    "index": tool_call.index,
                                    "function": {
                                        "name": tool_call.function_name,
                                        "arguments": tool_call.arguments,
                                    },
                                }
                            )
                    self._session.chat_history.append(history_item)
            if isinstance(item, FunctionCallFrame):  # send input for function call
                func_res = await asyncio.to_thread(
                    FunctionManager.execute,
                    item.function_name,
                    self._session,
                    **item.arguments_dict,
                )
                history_item = {
                    "role": "input",
                    "content": [
                        {"type": "text", "text": func_res},
                        {
                            "type": "text",
                            "text": "\n\n\n请用口语化形式总结检索结果，简短地回答用户的问题。",
                        },
                    ],
                }
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item["tool_call_id"] = item.tool_call_id
                self._session.chat_history.append(history_item)
                self._session.chat_history.append(
                    {
                        "role": "assistant",
                        "content": "<tts_start>",
                        "eot": False,
                    },  # Insert <tts_start> for speech response
                )
                self._session.ctx.state["messages"] = self._session.chat_history.to_list()
                self.send_input(self._session)
            yield item
        self._session.increment_chat_round()


# Chat: multi turn AQAA
class StepAudio2TextAudioChatProcessor(StepAudio2BaseProcessor):
    """
    Audio -> audio_LLM -> audio and text
    - A1-T2A2 (Audio Query and Audio+text Answer)
    - system prompt example:
        - AQAA: "You are a helpful assistant."
    """

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            audio = frame.path
        else:
            audio = bytes2TorchTensorWith16(frame.audio)

        if audio is None or len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        self._session.chat_history.append(
            {"role": "human", "content": [{"type": "audio", "audio": audio}]}
        )
        self._session.chat_history.append(
            {
                "role": "assistant",
                "content": "<tts_start>",
                "eot": False,
            },  # Insert <tts_start> for speech response
        )
        self._session.ctx.state["messages"] = self._session.chat_history.to_list()
        self.send_input(self._session)
        async for item in self.gen():
            if isinstance(item, LLMGenedTokensFrame):
                self._session.chat_history.pop(-1)
                if isinstance(self._audio_llm, TransformersManualVoiceStep2):
                    self._session.chat_history.append(
                        {
                            "role": "assistant",
                            "content": [
                                {"type": "text", "text": "<tts_start>"},
                                {"type": "token", "token": item.token_ids},
                            ],
                        }
                    )
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item = {
                        "role": "assistant",
                        "tts_content": {
                            "tts_text": item.text_tokens,
                            "tts_audio": item.audio_tokens,
                        },
                    }
                    if len(item.tool_calls) > 0:
                        history_item["tool_calls"] = []
                        for tool_call in item.tool_calls:
                            history_item["tool_calls"].append(
                                {
                                    "id": tool_call.tool_call_id,
                                    "type": tool_call.type,
                                    "index": tool_call.index,
                                    "function": {
                                        "name": tool_call.function_name,
                                        "arguments": tool_call.arguments,
                                    },
                                }
                            )
                    self._session.chat_history.append(history_item)
            if isinstance(item, FunctionCallFrame):  # send input for function call
                func_res = await asyncio.to_thread(
                    FunctionManager.execute,
                    item.function_name,
                    self._session,
                    **item.arguments_dict,
                )
                history_item = {
                    "role": "input",
                    "content": [
                        {"type": "text", "text": func_res},
                        {
                            "type": "text",
                            "text": "\n\n\n请用口语化形式总结检索结果，简短地回答用户的问题。",
                        },
                    ],
                }
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item["tool_call_id"] = item.tool_call_id
                self._session.chat_history.append(history_item)
                self._session.chat_history.append(
                    {
                        "role": "assistant",
                        "content": "<tts_start>",
                        "eot": False,
                    },  # Insert <tts_start> for speech response
                )
                self._session.ctx.state["messages"] = self._session.chat_history.to_list()
                self.send_input(self._session)
            yield item
        self._session.increment_chat_round()


# Chat: multi turn AQAA with reasoning think
class StepAudio2TextAudioThinkChatProcessor(StepAudio2BaseProcessor):
    SYS_PROMPT = "你的名字叫小跃，你是由阶跃星辰(StepFun)公司训练出来的语音大模型，你能听见用户的声音特征并在思维过程中描述出来，请激活深度思考模式，通过逐步分析、逻辑推理来解决用户的问题。"

    async def run_voice(self, frame: AudioRawFrame) -> AsyncGenerator[Frame, None]:
        if isinstance(frame, PathAudioRawFrame):
            audio = frame.path
        else:
            audio = bytes2TorchTensorWith16(frame.audio)

        if audio is None or len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        self._session.chat_history.append(
            {"role": "human", "content": [{"type": "audio", "audio": audio}]}
        )
        self._session.chat_history.append(
            {"role": "assistant", "content": "\n<think>\n", "eot": False}
        )
        self._session.ctx.state["messages"] = self._session.chat_history.to_list()
        # get think content, stop when "</think>" appears
        # if use hf transformers, stop_strings need text tokenizer
        if isinstance(self._audio_llm, TransformersManualVoiceStep2):
            self.send_input(self._session, stop_strings=["</think>"])
        elif isinstance(self._audio_llm, VllmClientStepAudio2):
            self.send_input(self._session, stop=["</think>"])
        else:
            raise Exception("Unsupported LLM engine type")
        async for item in self.gen():
            if isinstance(item, LLMGenedTokensFrame):
                self._session.chat_history.pop(-1)
                if isinstance(self._audio_llm, TransformersManualVoiceStep2):
                    self._session.chat_history.append(
                        {
                            "role": "assistant",
                            "content": [
                                {"type": "text", "text": "<tts_start>"},
                                {"type": "token", "token": item.token_ids},
                            ],
                        }
                    )
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item = {
                        "role": "assistant",
                        "tts_content": {
                            "tts_text": item.text_tokens,
                            "tts_audio": item.audio_tokens,
                        },
                    }
                    if len(item.tool_calls) > 0:
                        history_item["tool_calls"] = []
                        for tool_call in item.tool_calls:
                            history_item["tool_calls"].append(
                                {
                                    "id": tool_call.tool_call_id,
                                    "type": tool_call.type,
                                    "index": tool_call.index,
                                    "function": {
                                        "name": tool_call.function_name,
                                        "arguments": tool_call.arguments,
                                    },
                                }
                            )
                    self._session.chat_history.append(history_item)
            if isinstance(item, FunctionCallFrame):  # send input for function call
                func_res = await asyncio.to_thread(
                    FunctionManager.execute,
                    item.function_name,
                    self._session,
                    **item.arguments_dict,
                )
                history_item = {
                    "role": "input",
                    "content": [
                        {"type": "text", "text": func_res},
                        {
                            "type": "text",
                            "text": "\n\n\n请用口语化形式总结检索结果，简短地回答用户的问题。",
                        },
                    ],
                }
                if isinstance(self._audio_llm, VllmClientStepAudio2):
                    history_item["tool_call_id"] = item.tool_call_id
                self._session.chat_history.append(history_item)
                self._session.chat_history.append(
                    {
                        "role": "assistant",
                        "content": "<tts_start>",
                        "eot": False,
                    },  # Insert <tts_start> for speech response
                )
                self._session.ctx.state["messages"] = self._session.chat_history.to_list()
                self.send_input(self._session)
            if isinstance(item, ReasoningThinkTextFrame):  # send input for reasoning think result
                self._session.chat_history.pop(-1)
                self._session.chat_history.append(
                    {
                        "role": "assistant",
                        "content": f"\n<think>\n{item.text}\n<tts_start>",
                        "eot": False,
                    }
                )
                self._session.ctx.state["messages"] = self._session.chat_history.to_list()
                self.send_input(self._session)
            yield item
        self._session.increment_chat_round()


# ----------------------------------------------------------------------------------


# A1T1->T2
class StepAudioText2TextProcessor(StepAudio2BaseProcessor):
    """
    audio+text -> audio_LLM -> text
    - A1T1->T2 (Audio Understanding with Text Query)
    """

    def __init__(self, **kwargs):
        kwargs["is_speaking"] = False
        super().__init__(**kwargs)

    async def run_voice(self, frame: TextQuestionsAudioRawFrame) -> AsyncGenerator[Frame, None]:
        audio = bytes2TorchTensorWith16(frame.audio)

        if len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        messages = [
            {"role": "system", "content": self._system_prompt},
            {
                "role": "human",
                "content": [
                    {"type": "audio", "audio": audio},
                    {"type": "text", "text": frame.text},
                ],
            },
            {"role": "assistant", "content": None},
        ]
        self._session.ctx.state["messages"] = messages
        self.send_input(self._session)
        async for item in self.gen():
            yield item


# ----------------------------------------------------------------------------------


# A1T1->T2A2
class StepAudioText2TextAudioProcessor(StepAudio2BaseProcessor):
    """
    audio+text -> audio_LLM -> audio+text
    - A1T1->T2A2 (Audio Understanding with Text Query)
    """

    async def run_voice(self, frame: TextQuestionsAudioRawFrame) -> AsyncGenerator[Frame, None]:
        audio = bytes2TorchTensorWith16(frame.audio)

        if len(audio) == 0:
            yield ErrorFrame("No audio tokens extracted")
            return

        messages = [
            {"role": "system", "content": self._system_prompt},
            {
                "role": "human",
                "content": [
                    {"type": "audio", "audio": audio},
                    {"type": "text", "text": frame.text},
                ],
            },
            {
                "role": "assistant",
                "content": "<tts_start>",
                "eot": False,
            },  # Insert <tts_start> for speech response
        ]
        self._session.ctx.state["messages"] = messages
        self.send_input(self._session)
        async for item in self.gen():
            yield item


# audio understanding with text query
class StepMMAUProcessor(StepAudioText2TextAudioProcessor):
    SYS_PROMPT = "You are an expert in audio analysis, please analyze the audio content and answer the questions accurately."


class StepMockAQAAProcessor(StepAudio2TextAudioChatProcessor):
    def __init__(self, token2wav=None, **kwargs):
        super().__init__(token2wav=token2wav, **kwargs)

    def put_out_audio_text(self, token_iter, is_out_text: bool = True, is_think: bool = False):
        wav_files = glob.glob(os.path.join(RECORDS_DIR, "bot_speak*.wav"))

        wav_files.sort(key=os.path.getmtime)

        for wav_file in wav_files:
            if self._generate_over is True:
                break
            try:
                with open(wav_file, "rb") as f:
                    self._token2wav.stream(
                        self._token2wav.WARMUP_TOKENS[
                            : self._token2wav.CHUNK_SIZE + self._token2wav.flow.pre_lookahead_len
                        ],
                        prompt_wav=self._prompt_wav,
                    )

                    frame = AudioRawFrame(
                        audio=f.read(),
                        sample_rate=self._audio_llm.RATE,
                        num_channels=1,
                    )
                    self._queue.put(frame)
            except Exception as e:
                print(f"Error reading {wav_file}: {str(e)}")

        self._queue.put(LLMGenedTokensFrame())

        return [], "", False


"""
python -m src.processors.voice.step_audio2_processor
"""
if __name__ == "__main__":
    from pathlib import Path

    import wave
    from apipeline.frames import AudioRawFrame, StartFrame, EndFrame, CancelFrame

    from achatbot.common.logger import Logger
    from achatbot.types.frames import PathAudioRawFrame
    from achatbot.cmd.bots.voice.step_audio2.helper import (
        get_step_audio2_processor,
    )
    from achatbot.types.ai_conf import AIConfig, LLMConfig

    Logger.init(os.getenv("LOG_LEVEL", "info").upper(), is_file=False, is_console=True)

    async def run_aqaa():
        processor = get_step_audio2_processor(
            LLMConfig(
                processor="StepAudio2TextAudioChatProcessor",
                args={
                    "init_system_prompt": "",
                    "prompt_wav": "./assets/default_male.wav",
                    "warmup_cn": 2,
                    "chat_history_size": None,
                    "text_stream_out": False,
                    "no_stream_sleep_time": 0.5,
                    "lm_model_name_or_path": "./models/stepfun-ai/Step-Audio-2-mini",
                },
            )
        )
        await processor.start(StartFrame())
        for round_idx, audio_path in enumerate(
            [
                "./deps/StepAudio2/assets/multi-turn-round1-听说荡口古镇从下个月开始取消门票了，你知道这事吗。.wav",
                "./deps/StepAudio2/assets/multi-turn-round2-新闻说九月十九号就免费开放了。好像整个古镇都升级改造了，现在变成开放式街区了。.wav",
            ]
        ):
            print("round: ", round_idx)
            frame_iter = processor.run_voice(
                PathAudioRawFrame(
                    path=audio_path,
                    audio=b"",
                )
            )
            audio = b""
            async for frame in frame_iter:
                if isinstance(frame, AudioRawFrame):
                    audio += frame.audio
                print(f"{round_idx=} gen_frame-->", frame)
            print(f"{round_idx=} {processor._session.chat_history=}")
            wav_path = Path(
                f"{ASSETS_DIR}/StepAudio2/output-processor-chunks-stream-{round_idx}.wav"
            )
            with wave.open(str(wav_path), "wb") as wf:
                wf.setnchannels(1)
                wf.setsampwidth(2)
                wf.setframerate(24000)
                wf.writeframes(audio)
        await processor.stop(EndFrame())

    asyncio.run(run_aqaa())



from collections import deque
import logging
import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, Optional

from apipeline.frames.base import Frame
from apipeline.frames.control_frames import StartFrame
from apipeline.frames.sys_frames import (
    StartInterruptionFrame,
    StopInterruptionFrame,
    SystemFrame,
    CancelFrame,
)
from apipeline.frames.data_frames import AudioRawFrame
from apipeline.processors.frame_processor import FrameDirection
from apipeline.processors.input_processor import InputProcessor

from src.common.ringbuffer import RingBuffer
from src.common.interface import IVADAnalyzer
from src.common.types import AudioVADTurnParams, VADState
from src.types.frames.control_frames import UserStartedSpeakingFrame, UserStoppedSpeakingFrame
from src.types.frames.sys_frames import BotInterruptionFrame, MetricsFrame
from src.types.speech.turn_analyzer import EndOfTurnState
from src.types.frames.data_frames import VADStateAudioRawFrame


class AudioVADInputProcessor(InputProcessor):
    def __init__(
        self,
        params: AudioVADTurnParams,
        name: str | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
        **kwargs,
    ):
        super().__init__(name=name, loop=loop, **kwargs)
        self._params = params
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._bytes_buffer = RingBuffer(
            self._params.audio_in_buffer_secs
            * self._params.audio_in_sample_rate
            * self._params.audio_in_channels
            * self._params.audio_in_sample_width
        )

        self._vad_analyzer: IVADAnalyzer | None = params.vad_analyzer
        self._num_bytes_required = 0
        if self._vad_analyzer:
            self._num_bytes_required = (
                self._vad_analyzer.num_frames_required()
                * self._params.audio_in_channels
                * self._params.audio_in_sample_width
            )

    @property
    def vad_analyzer(self) -> IVADAnalyzer | None:
        return self._params.vad_analyzer

    def set_vad_analyzer(self, analyzer: IVADAnalyzer):
        self._vad_analyzer = analyzer

    #
    # Audio task
    #

    async def start(self, frame: StartFrame):
        logging.info(f"{self.name} start, params: {self._params}")
        # Create audio input queue and task if needed.
        if self._params.audio_in_enabled or self._params.vad_enabled:
            self._audio_in_queue = asyncio.Queue()
            self._audio_task = self.get_event_loop().create_task(self._audio_task_handler())

    async def stop(self):
        # Wait for the task to finish.
        if self._params.audio_in_enabled or self._params.vad_enabled:
            logging.info("stop to Cancelling audio task")
            self._audio_task.cancel()
            await self._audio_task

        # Wait for the push frame task to finish. It will finish when the
        # EndFrame is actually processed. wait timeout 1s
        try:
            self._push_frame_task and await asyncio.wait_for(self._push_frame_task, 1)
        except asyncio.TimeoutError:
            pass
        except asyncio.CancelledError:
            pass

    async def cancel(self, frame: CancelFrame):
        """
        Cancel all the tasks and wait for them to finish.
        """
        if self._params.audio_in_enabled or self._params.vad_enabled:
            logging.info("cancel to Cancelling audio task")
            # Wait for audio_task cancel to finish
            self._audio_task.cancel()
            await self._audio_task

        # Wait for async processor push_frame_task cancel to finish
        await self.cleanup()

    async def push_audio_frame(self, frame: AudioRawFrame):
        if self._params.audio_in_enabled or self._params.vad_enabled:
            await self._audio_in_queue.put(frame)

    async def _audio_task_handler(self):
        vad_state: VADState = VADState.QUIET
        while True:
            try:
                frame: AudioRawFrame = await asyncio.wait_for(self._audio_in_queue.get(), timeout=1)

                if self._num_bytes_required == 0 and self._params.vad_audio_passthrough:
                    # Note: no VAD, just push the frame downstream. e.g. use moshi/gemini live BIDI model
                    await self.push_frame(frame)
                    continue

                self._bytes_buffer.extend_bytes(frame.audio)
                while self._bytes_buffer.size() >= self._num_bytes_required:
                    chunk_bytes = self._bytes_buffer.pop_bytes(self._num_bytes_required)
                    previous_vad_state = vad_state
                    # Check VAD and push event if necessary. We just care about
                    # changes from QUIET to SPEAKING and vice versa.
                    if self._params.vad_enabled:
                        vad_state_frame, user_interuption_frame = await self._handle_vad(
                            chunk_bytes, vad_state
                        )
                        vad_state = vad_state_frame.state

                    if user_interuption_frame and isinstance(
                        user_interuption_frame, UserStartedSpeakingFrame
                    ):  # start speak
                        await self.handle_interruptions(user_interuption_frame, True)

                    # Push audio downstream if passthrough.
                    if self._params.vad_enabled and self._params.vad_audio_passthrough:
                        if len(vad_state_frame.audio) > 0:
                            await self.push_frame(vad_state_frame)
                    else:
                        await self.push_frame(frame)

                    if user_interuption_frame and isinstance(
                        user_interuption_frame, UserStoppedSpeakingFrame
                    ):  # stop speak without turn analyzer
                        await self.handle_interruptions(user_interuption_frame, True)

                    if self._params.turn_analyzer:  # stop speak with turn analyzer
                        await self._run_turn_analyzer(frame, vad_state, previous_vad_state)

            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                logging.info(f"{self.name} audio_task_handler cancelled")
                break
            except Exception as e:
                logging.exception(f"{self} error reading audio frames: {e}")
                if self.get_event_loop().is_closed():
                    logging.warning(f"{self.name} event loop is closed")
                    break
                # Handle RuntimeError for shutdown executor
                if isinstance(
                    e, RuntimeError
                ) and "cannot schedule new futures after shutdown" in str(e):
                    logging.warning(f"{self.name} executor shutdown, stopping audio task")
                    break

    async def _vad_analyze(self, audio_bytes: bytes) -> VADStateAudioRawFrame:
        vad_state_frame = VADStateAudioRawFrame(state=VADState.QUIET, audio=audio_bytes)
        if self.vad_analyzer:
            vad_state_frame: VADStateAudioRawFrame = await self.get_event_loop().run_in_executor(
                self._executor, self.vad_analyzer.analyze_audio, audio_bytes
            )
        return vad_state_frame

    async def _handle_vad(self, audio_bytes: bytes, vad_state: VADState) -> VADStateAudioRawFrame:
        vad_state_frame = await self._vad_analyze(audio_bytes)
        new_vad_state = vad_state_frame.state
        user_interuption_frame = None
        if (
            new_vad_state != vad_state
            and new_vad_state != VADState.STARTING
            and new_vad_state != VADState.STOPPING
        ):
            can_create_user_frames = (
                self._params.turn_analyzer is None
                or not self._params.turn_analyzer.speech_triggered
            )
            if new_vad_state == VADState.SPEAKING:
                if can_create_user_frames:
                    user_interuption_frame = UserStartedSpeakingFrame()
            elif new_vad_state == VADState.QUIET:
                if can_create_user_frames:
                    user_interuption_frame = UserStoppedSpeakingFrame()

            vad_state_frame.state = new_vad_state

        return vad_state_frame, user_interuption_frame

    async def _run_turn_analyzer(
        self, frame: AudioRawFrame, vad_state: VADState, previous_vad_state: VADState
    ):
        """Run turn analysis on audio frame and handle results."""
        is_speech = vad_state == VADState.SPEAKING or vad_state == VADState.STARTING
        # If silence exceeds threshold, we are going to receive EndOfTurnState.COMPLETE
        end_of_turn_state = self._params.turn_analyzer.append_audio(frame.audio, is_speech)
        if end_of_turn_state == EndOfTurnState.COMPLETE:
            await self._handle_end_of_turn_complete(end_of_turn_state)
        # Otherwise we are going to trigger to check if the turn is completed based on the VAD
        elif vad_state == VADState.QUIET and vad_state != previous_vad_state:
            await self._handle_end_of_turn()

    async def _handle_end_of_turn_complete(self, state: EndOfTurnState):
        """Handle completion of end-of-turn analysis."""
        if state == EndOfTurnState.COMPLETE:
            await self.handle_interruptions(UserStoppedSpeakingFrame(), True)

    async def _handle_end_of_turn(self):
        """Handle end-of-turn analysis and generate prediction results."""
        if self._params.turn_analyzer:
            state, prediction = await self._params.turn_analyzer.analyze_end_of_turn()
            await self._handle_prediction_result(prediction)
            await self._handle_end_of_turn_complete(state)

    async def _handle_prediction_result(self, result: Optional[Dict[str, Any]] = None):
        """Handle a prediction result event from the turn analyzer."""
        pass
        # todo: add metrics frame to push

    #
    # Handle interruptions
    #
    async def _start_interruption(self):
        if not self.interruptions_allowed:
            return

        # use async frame processor handle_interruptions
        # (cancel old then create new)
        await super()._handle_interruptions(StartInterruptionFrame())

    async def _stop_interruption(self):
        if not self.interruptions_allowed:
            return

        await self.push_frame(StopInterruptionFrame())

    async def handle_interruptions(self, frame: Frame, push_frame: bool):
        if self.interruptions_allowed:
            # Make sure we notify about interruptions quickly out-of-band
            if isinstance(frame, BotInterruptionFrame):
                logging.debug("Bot interruption")
                await self._start_interruption()
            elif isinstance(frame, UserStartedSpeakingFrame):
                logging.info("User started speaking")
                await self._start_interruption()
            elif isinstance(frame, UserStoppedSpeakingFrame):
                logging.info("User stopped speaking")
                await self._stop_interruption()

        if push_frame:
            await self.push_frame(frame)

    #
    # Process frame
    #

    async def process_sys_frame(self, frame: Frame, direction: FrameDirection):
        if isinstance(frame, BotInterruptionFrame):
            await self.handle_interruptions(frame, False)
        elif isinstance(frame, StartInterruptionFrame):
            await self._start_interruption()
        elif isinstance(frame, StopInterruptionFrame):
            await self._stop_interruption()
        # All other system frames
        elif isinstance(frame, SystemFrame):
            await self.push_frame(frame, direction)

    async def process_frame(self, frame: Frame, direction: FrameDirection):
        return await super().process_frame(frame, direction)



import os
from pathlib import Path

import modal

from .modal_webrtc_signaling_server import ModalWebRtcSignalingServer
from .bot_webrtc_peer import BotWebRtcPeer, app


base_image = (
    modal.Image.debian_slim(python_version="3.12")
    .apt_install("python3-opencv", "ffmpeg")
    .pip_install(
        "fastapi[standard]==0.115.4",
        "aiortc==1.11.0",
        "opencv-python==4.11.0.86",
        "shortuuid==1.0.13",
    )
)

this_directory = Path(__file__).parent.resolve()

server_image = base_image.add_local_dir(this_directory / "frontend", remote_path="/frontend")


@app.cls(
    image=server_image,
    max_containers=1,
)
class WebcamWebRtcSignalingServer(ModalWebRtcSignalingServer):
    def get_modal_peer_class(self):
        return BotWebRtcPeer

    def initialize(self):
        from fastapi.responses import HTMLResponse
        from fastapi.staticfiles import StaticFiles

        self.web_app.mount("/static", StaticFiles(directory="/frontend"))

        @self.web_app.get("/")
        async def root():
            html = open("/frontend/index.html").read()
            return HTMLResponse(content=html)

        print("----initialized singaling service----")



import logging
import json
from collections.abc import Callable

from a2a.client import (
    Client,
    ClientFactory,
)
from a2a.types import (
    AgentCard,
    Message,
    Task,
    TaskArtifactUpdateEvent,
    TaskState,
    TaskStatusUpdateEvent,
)


TaskCallbackArg = Task | TaskStatusUpdateEvent | TaskArtifactUpdateEvent
TaskUpdateCallback = Callable[[TaskCallbackArg, AgentCard], Task]


class RemoteAgentConnections:
    """A class to hold the connections to the remote agents."""

    def __init__(self, client_factory: ClientFactory, agent_card: AgentCard):
        self.agent_client: Client = client_factory.create(agent_card)
        self.card: AgentCard = agent_card
        self.pending_tasks = set()

    def get_agent(self) -> AgentCard:
        return self.card

    async def send_message(self, message: Message) -> Task | Message | None:
        lastTask: Task | None = None
        try:
            async for event in self.agent_client.send_message(message):
                if isinstance(event, Message):
                    return event
                if isinstance(event, tuple) and self.is_terminal_or_interrupted(event[0]):
                    return event[0]
                lastTask = event[0]
        except Exception as e:
            logging.exception("Exception found in send_message")
            raise e
        return lastTask

    def is_terminal_or_interrupted(self, task: Task) -> bool:
        return task.status.state in [
            TaskState.completed,
            TaskState.canceled,
            TaskState.failed,
            TaskState.input_required,
            TaskState.unknown,
        ]


"""
python -m src.services.a2a_multiagents.remote_agent_connection
"""
if __name__ == "__main__":
    import asyncio
    import uuid

    import httpx
    from a2a.client import A2ACardResolver, ClientConfig, ClientFactory
    from a2a.types import (
        AgentCard,
        TransportProtocol,
    )
    from a2a.utils.message import new_agent_text_message

    async def a_run():
        async with httpx.AsyncClient(timeout=60) as httpx_client:
            config = ClientConfig(
                httpx_client=httpx_client,
                supported_transports=[
                    TransportProtocol.jsonrpc,
                    TransportProtocol.http_json,
                    # TransportProtocol.grpc,
                ],
            )
            client_factory = ClientFactory(config)
            card_resolver = A2ACardResolver(httpx_client, "http://0.0.0.0:6666")
            card = await card_resolver.get_agent_card()
            conn = RemoteAgentConnections(client_factory, card)

            message = new_agent_text_message(
                text="what's a chatbot",
                context_id=str(uuid.uuid4()),
                # task_id=str(uuid.uuid4()),# new message no task_id
            )
            res = await conn.send_message(message)
            print(json.dumps(res.model_dump_json()))
            # print("res-->", res)
            await asyncio.sleep(1)

    asyncio.run(a_run())



import json
import logging
import math
from pathlib import Path
import re
import sys
from threading import Lock
from typing import AsyncGenerator
import os
import io

import numpy as np
import torch
import torchaudio


try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../../StepAudio"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../../deps/StepAudio"))
    from achatbot.core.llm.transformers.manual_speech_step import TransformersManualSpeechStep
    from achatbot.StepAudio.tokenizer import StepAudioTokenizer
    from achatbot.StepAudio.cosyvoice.cli.cosyvoice import CosyVoice
except ModuleNotFoundError as e:
    logging.error("In order to use step-tts, you need to `pip install achatbot[tts_step]`.")
    raise Exception(f"Missing module: {e}")

from achatbot.common.random import set_all_random_seed
from achatbot.common.types import PYAUDIO_PAFLOAT32, ASSETS_DIR
from achatbot.common.interface import ITts
from achatbot.common.session import Session
from achatbot.types.speech.tts.step import StepTTSArgs
from achatbot.types.llm.transformers import TransformersSpeechLMArgs
from .base import BaseTTS


class StepTTS(BaseTTS, ITts):
    """
    support tts mode:
    - lm_gen: text+ref audio waveform lm gen audio wav code to gen waveform with static batch stream:
        text+ref audio waveform -> tokenizer -> text+audio token ids -> step1 lm  -> audio token ids (wav_code) -> flow(CFM) -> mel - vocoder(hifi) -> waveform
    - voice_clone: voice clone w/o lm gen, decode wav code:
        src+ref audio waveform -> speech tokenizer-> audio token ids (wav_code) -> flow(CFM) -> mel - vocoder(hifi) -> clone ref audio waveform
    """

    TAG = "tts_step"

    def __init__(self, **kwargs) -> None:
        self.args = StepTTSArgs(**kwargs)
        assert (
            self.args.stream_factor >= 2
        ), "stream_factor must >=2 increase for better speech quality, but rtf slow (speech quality vs rtf)"

        self.encoder = StepAudioTokenizer(self.args.speech_tokenizer_model_path)

        self.lm_args = TransformersSpeechLMArgs(**self.args.lm_args)
        self.lm_model = TransformersManualSpeechStep(**self.lm_args.__dict__)
        # session ctx dict with lock, maybe need a session class
        self.session_lm_generat_lock = Lock()
        self.session_lm_generated_ids = {}  # session_id: ids(ptr)

        self.common_cosy_model = CosyVoice(
            os.path.join(self.lm_args.lm_model_name_or_path, "CosyVoice-300M-25Hz")
        )
        self.music_cosy_model = CosyVoice(
            os.path.join(self.lm_args.lm_model_name_or_path, "CosyVoice-300M-25Hz-Music")
        )

        self.sys_prompt_dict = {
            "sys_prompt_for_rap": "请参考对话历史里的音色，用RAP方式将文本内容大声说唱出来。",
            "sys_prompt_for_vocal": "请参考对话历史里的音色，用哼唱的方式将文本内容大声唱出来。",
            "sys_prompt_wo_spk": '作为一名卓越的声优演员，你的任务是根据文本中（）或()括号内标注的情感、语种或方言、音乐哼唱、语音调整等标签，以丰富细腻的情感和自然顺畅的语调来朗读文本。\n# 情感标签涵盖了多种情绪状态，包括但不限于：\n- "高兴1"\n- "高兴2"\n- "生气1"\n- "生气2"\n- "悲伤1"\n- "撒娇1"\n\n# 语种或方言标签包含多种语言或方言，包括但不限于：\n- "中文"\n- "英文"\n- "韩语"\n- "日语"\n- "四川话"\n- "粤语"\n- "广东话"\n\n# 音乐哼唱标签包含多种类型歌曲哼唱，包括但不限于：\n- "RAP"\n- "哼唱"\n\n# 语音调整标签，包括但不限于：\n- "慢速1"\n- "慢速2"\n- "快速1"\n- "快速2"\n\n请在朗读时，根据这些情感标签的指示，调整你的情感、语气、语调和哼唱节奏，以确保文本的情感和意义得到准确而生动的传达，如果没有()或（）括号，则根据文本语义内容自由演绎。',
            "sys_prompt_with_spk": '作为一名卓越的声优演员，你的任务是根据文本中（）或()括号内标注的情感、语种或方言、音乐哼唱、语音调整等标签，以丰富细腻的情感和自然顺畅的语调来朗读文本。\n# 情感标签涵盖了多种情绪状态，包括但不限于：\n- "高兴1"\n- "高兴2"\n- "生气1"\n- "生气2"\n- "悲伤1"\n- "撒娇1"\n\n# 语种或方言标签包含多种语言或方言，包括但不限于：\n- "中文"\n- "英文"\n- "韩语"\n- "日语"\n- "四川话"\n- "粤语"\n- "广东话"\n\n# 音乐哼唱标签包含多种类型歌曲哼唱，包括但不限于：\n- "RAP"\n- "哼唱"\n\n# 语音调整标签，包括但不限于：\n- "慢速1"\n- "慢速2"\n- "快速1"\n- "快速2"\n\n请在朗读时，使用[{}]的声音，根据这些情感标签的指示，调整你的情感、语气、语调和哼唱节奏，以确保文本的情感和意义得到准确而生动的传达，如果没有()或（）括号，则根据文本语义内容自由演绎。',
        }

        self.speakers_info = {}
        self.register_speakers()

        # lm model gen warmup, codec model decode(flow + hifi) don't to warmup

    def register_speakers(self):
        self.speakers_info = {}

        speackers_info_path = os.path.join(ASSETS_DIR, "speakers/speakers_info.json")
        with open(speackers_info_path, "r") as f:
            speakers_info = json.load(f)

        for speaker_id, prompt_text in speakers_info.items():
            prompt_wav_path = os.path.join(ASSETS_DIR, f"speakers/{speaker_id}_prompt.wav")
            (
                ref_audio_code,
                ref_audio_token,
                ref_audio_token_len,
                ref_speech_feat,
                ref_speech_feat_len,
                ref_speech_embedding,
            ) = self.preprocess_prompt_wav(prompt_wav_path)

            self.speakers_info[speaker_id] = {
                "ref_text": prompt_text,
                "ref_audio_code": ref_audio_code,
                "ref_speech_feat": ref_speech_feat.to(torch.bfloat16),
                "ref_speech_feat_len": ref_speech_feat_len,
                "ref_speech_embedding": ref_speech_embedding.to(torch.bfloat16),
                "ref_audio_token": ref_audio_token,
                "ref_audio_token_len": ref_audio_token_len,
            }
            logging.info(f"Registered speaker: {speaker_id}")

    def wav2code(self, prompt_wav_path: str):
        prompt_wav, prompt_wav_sr = torchaudio.load(prompt_wav_path)
        if prompt_wav.shape[0] > 1:
            prompt_wav = prompt_wav.mean(dim=0, keepdim=True)  # multi-channel to mono
        prompt_code, _, _ = self.encoder.wav2token(prompt_wav, prompt_wav_sr)
        return prompt_code

    def preprocess_prompt_wav(self, prompt_wav_path: str):
        prompt_wav, prompt_wav_sr = torchaudio.load(prompt_wav_path)
        if prompt_wav.shape[0] > 1:
            prompt_wav = prompt_wav.mean(dim=0, keepdim=True)  # multi-channel to mono
        prompt_wav_16k = torchaudio.transforms.Resample(orig_freq=prompt_wav_sr, new_freq=16000)(
            prompt_wav
        )
        prompt_wav_22k = torchaudio.transforms.Resample(orig_freq=prompt_wav_sr, new_freq=22050)(
            prompt_wav
        )

        speech_feat, speech_feat_len = self.common_cosy_model.frontend._extract_speech_feat(
            prompt_wav_22k
        )
        speech_embedding = self.common_cosy_model.frontend._extract_spk_embedding(prompt_wav_16k)

        prompt_code, _, _ = self.encoder.wav2token(prompt_wav, prompt_wav_sr)
        prompt_token = torch.tensor([prompt_code], dtype=torch.long) - 65536
        prompt_token_len = torch.tensor([prompt_token.shape[1]], dtype=torch.long)

        return (
            prompt_code,
            prompt_token,
            prompt_token_len,
            speech_feat,
            speech_feat_len,
            speech_embedding,
        )

    def set_voice(self, ref_audio_path: str, **kwargs):
        """
        - save to speacker info dict
        TODO: save dict to dist kv store
        """
        assert os.path.exists(ref_audio_path), "ref_audio_path is not exists"
        assert kwargs.get(
            "ref_speaker", None
        ), "ref_speaker is not exists"  # maybe use random speaker
        assert kwargs.get(
            "ref_text", None
        ), "ref_text is not exists"  # maybe use asr to get ref_text
        ref_speaker = kwargs.get("ref_speaker")
        ref_text = kwargs.get("ref_text")

        (
            ref_audio_code,
            ref_audio_token,
            ref_audio_token_len,
            ref_speech_feat,
            ref_speech_feat_len,
            ref_speech_embedding,
        ) = self.preprocess_prompt_wav(ref_audio_path)

        self.speakers_info[ref_speaker] = {
            "ref_text": ref_text,
            "ref_audio_code": ref_audio_code,
            "ref_speech_feat": ref_speech_feat.to(torch.bfloat16),
            "ref_speech_feat_len": ref_speech_feat_len,
            "ref_speech_embedding": ref_speech_embedding.to(torch.bfloat16),
            "ref_audio_token": ref_audio_token,
            "ref_audio_token_len": ref_audio_token_len,
        }

    def get_voices(self) -> list:
        return list(self.speakers_info.keys())

    def get_stream_info(self) -> dict:
        return {
            # "format": PYAUDIO_PAINT16,
            "format": PYAUDIO_PAFLOAT32,
            "channels": 1,
            "rate": 22050,
            "sample_width": 2,
            # "np_dtype": np.int16,
            "np_dtype": np.float32,
        }

    def detect_instruction_name(self, text):
        instruction_name = ""
        match_group = re.match(r"^([（\(][^\(\)()]*[）\)]).*$", text, re.DOTALL)
        if match_group is not None:
            instruction = match_group.group(1)
            instruction_name = instruction.strip("()（）")
        return instruction_name

    def set_system_prompt(self, text, ref_speaker: str = "Tingting"):
        sys_prompt = self.sys_prompt_dict["sys_prompt_wo_spk"]
        instruction_name = self.detect_instruction_name(text)
        if instruction_name:
            if "哼唱" in text:
                sys_prompt = self.sys_prompt_dict["sys_prompt_for_vocal"]
            else:
                sys_prompt = self.sys_prompt_dict["sys_prompt_for_rap"]
        elif ref_speaker:
            sys_prompt = self.sys_prompt_dict["sys_prompt_with_spk"].format(ref_speaker)
        self.lm_model.set_system_prompt(sys_prompt=sys_prompt)

    def voice_clone(self, session: Session, ref_speaker: str, cosy_model, **kwargs):
        """
        - voice_clone: voice clone w/o lm gen, decode wav code:
        src+ref audio waveform -> speech tokenizer-> audio token ids (wav_code) -> flow(CFM) -> mel - vocoder(hifi) -> clone ref audio waveform
        """
        src_audio_path = kwargs.get("src_audio_path", None)
        if not src_audio_path or not os.path.exists(src_audio_path):
            logging.error(f"{src_audio_path} is not exists")
            return None

        src_audio_code = self.wav2code(src_audio_path)
        tensor_audio_token_ids = torch.tensor([src_audio_code]).to(torch.long).to("cuda") - 65536
        tts_speech = cosy_model.token_to_wav_offline(
            tensor_audio_token_ids,
            self.speakers_info[ref_speaker]["ref_speech_feat"].to(torch.bfloat16),
            self.speakers_info[ref_speaker]["ref_speech_feat_len"],
            self.speakers_info[ref_speaker]["ref_audio_token"],
            self.speakers_info[ref_speaker]["ref_audio_token_len"],
            self.speakers_info[ref_speaker]["ref_speech_embedding"].to(torch.bfloat16),
        )
        return tts_speech.float().numpy().tobytes()

    async def lm_gen(
        self,
        session: Session,
        text: str,
        ref_speaker: str,
        batch_size: int,
        cosy_model,
        **kwargs,
    ) -> AsyncGenerator[bytes, None]:
        """
        - lm_gen: text+ref audio waveform lm gen audio wav code to gen waveform with static batch stream:
        text+ref audio waveform -> tokenizer -> text+audio token ids -> step1 lm -> audio token ids (wav_code) -> flow(CFM) -> mel - vocoder(hifi) -> waveform
        """
        session_id = session.ctx.client_id

        self.set_system_prompt(text, ref_speaker=ref_speaker)

        one_shot_ref_text = self.speakers_info[ref_speaker]["ref_text"]
        one_shot_ref_audio = self._tokenizer.decode(
            self.speakers_info[ref_speaker]["ref_audio_code"]
        )
        prompt = f"<s><|BOT|><s> system\n{self.sys_prompt}"
        prompt += f"<|EOT|><|BOT|><s> human\n{one_shot_ref_text}" if one_shot_ref_text else ""
        prompt += f"<|EOT|><|BOT|><s> assistant\n{one_shot_ref_audio}" if one_shot_ref_audio else ""
        prompt += f"<|EOT|><|BOT|><s> human\n{text}"
        prompt += "<|EOT|><|BOT|><s> assistant\n"

        session.ctx.state["prompt"] = prompt
        audio_vq_tokens = self.lm_model.generate(session, **kwargs)
        for token_id in audio_vq_tokens:
            # print(token_id, end=",", flush=True)
            if token_id == self.lm_model.end_token_id:  # skip <|EOT|>, break
                break
            self.session_lm_generated_ids[session_id].append(token_id)
            if len(self.session_lm_generated_ids[session_id]) % batch_size == 0:
                batch = (
                    torch.tensor(self.session_lm_generated_ids[session_id])
                    .unsqueeze(0)
                    .to(cosy_model.model.device)
                    - 65536
                )  # [T] -> [1,T]
                logging.debug(f"batch: {batch}")
                # Process each batch
                sub_tts_speech = cosy_model.token_to_wav_offline(
                    batch,
                    self.speakers_info[ref_speaker]["ref_speech_feat"].to(torch.bfloat16),
                    self.speakers_info[ref_speaker]["ref_speech_feat_len"],
                    self.speakers_info[ref_speaker]["ref_audio_token"],
                    self.speakers_info[ref_speaker]["ref_audio_token_len"],
                    self.speakers_info[ref_speaker]["ref_speech_embedding"].to(torch.bfloat16),
                )
                yield sub_tts_speech.float().numpy().tobytes()
                with self.session_lm_generat_lock:
                    self.session_lm_generated_ids[session_id] = []

        if len(self.session_lm_generated_ids[session_id]) > 0:
            batch = (
                torch.tensor(self.session_lm_generated_ids[session_id])
                .unsqueeze(0)
                .to(cosy_model.model.device)
                - 65536
            )  # [T] -> [1,T]
            logging.debug(f"batch: {batch}")
            # Process each batch
            sub_tts_speech = cosy_model.token_to_wav_offline(
                batch,
                self.speakers_info[ref_speaker]["ref_speech_feat"].to(torch.bfloat16),
                self.speakers_info[ref_speaker]["ref_speech_feat_len"],
                self.speakers_info[ref_speaker]["ref_audio_token"],
                self.speakers_info[ref_speaker]["ref_audio_token_len"],
                self.speakers_info[ref_speaker]["ref_speech_embedding"].to(torch.bfloat16),
            )
            yield sub_tts_speech.float().numpy().tobytes()

    async def _inference(
        self, session: Session, text: str, **kwargs
    ) -> AsyncGenerator[bytes, None]:
        if "cuda" in str(self.lm_model._model.device):
            torch.cuda.empty_cache()
        seed = kwargs.get("seed", self.lm_args.lm_gen_seed)
        set_all_random_seed(seed)

        ref_speaker = kwargs.pop("ref_speaker", "Tingting")
        instruction_name = self.detect_instruction_name(text)
        cosy_model = self.common_cosy_model
        if instruction_name in ["RAP", "哼唱"]:
            cosy_model = self.music_cosy_model
            ref_speaker = f"{ref_speaker}{instruction_name}"
            if ref_speaker not in self.speakers_info:
                ref_speaker = f"Tingting{instruction_name}"
        if ref_speaker and ref_speaker not in self.speakers_info:
            ref_speaker = "Tingting"
        logging.debug(f"use ref_speaker: {ref_speaker}")

        assert (
            kwargs.get("stream_factor", self.args.stream_factor) >= 2
        ), "stream_factor must >=2 increase for better speech quality, but rtf slow (speech quality vs rtf)"
        batch_size = math.ceil(
            kwargs.get("stream_factor", self.args.stream_factor)
            * cosy_model.model.flow.input_frame_rate
        )

        session_id = session.ctx.client_id
        with self.session_lm_generat_lock:
            self.session_lm_generated_ids[session_id] = []

        tts_mode = kwargs.get("tts_mode", self.args.tts_mode)
        if tts_mode == "voice_clone":
            tts_speech = self.voice_clone(session, ref_speaker, cosy_model, **kwargs)
            yield tts_speech
        else:  # lm_gen
            async for item in self.lm_gen(
                session, text, ref_speaker, batch_size, cosy_model, **kwargs
            ):
                yield item

        with self.session_lm_generat_lock:
            self.session_lm_generated_ids.pop(session_id)
        torch.cuda.empty_cache()



from dataclasses import dataclass, field

from src.types.llm.transformers import TransformersLMArgs


@dataclass
class Qwen3OmniCode2WavArgs:
    """Streaming Codec Decoder"""

    chunk_size: int = 50
    left_context_size: int = 25


@dataclass
class Qwen3TransformersVisionVoiceLMArgs(TransformersLMArgs):
    """
    text+vision(Image/video)+voice(audio+speech) lm args (thinker + talker) + code2wav(Streaming Codec Decoder) args
    """

    thinker_eos_token_ids: list = field(default_factory=lambda: [151643, 151645])
    thinker_stop_strings_per_step: list = field(default_factory=lambda: [".", "。"])
    thinker_args: dict = field(default_factory=lambda: TransformersLMArgs().__dict__)
    talker_args: dict = field(default_factory=lambda: TransformersLMArgs().__dict__)
    talker_skip_thinker_token_ids: list[int] = field(default_factory=lambda: [])
    talker_eos_token_ids: list[int] = field(default_factory=lambda: [8292, 8294])
    speaker: str = "Chelsie"
    disable_talker: bool = False
    code2wav_args: dict = field(default_factory=lambda: Qwen3OmniCode2WavArgs().__dict__)



# Adopted from https://github.com/lm-sys/FastChat. Below is the original copyright:
# Adopted from tatsu-lab@stanford_alpaca. Below is the original copyright:
#    Copyright 2023 Rohan Taori, Ishaan Gulrajani, Tianyi Zhang, Yann Dubois, Xuechen Li
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import pathlib
import torch
import transformers

# from GOT.train.trainer import GOTTrainer
# from GOT.train.trainer_vit_llrd import GOTTrainer
from GOT.train.trainer_vit_fixlr import GOTTrainer
from GOT.model import GOTLlamaForCausalLM
from GOT.data import make_supervised_data_module
from GOT.utils.arguments import *
from GOT.utils.constants import *
from GOT.utils.utils import *


# def find_all_linear_names(model):
#     cls = torch.nn.Linear
#     lora_module_names = set()
#     for name, module in model.named_modules():
#         if isinstance(module, cls):
#             names = name.split('.')
#             lora_module_names.add(names[0] if len(names) == 1 else names[-1])


#     if 'lm_head' in lora_module_names: # needed for 16-bit
#         lora_module_names.remove('lm_head')
#     return list(lora_module_names)

def train():
    parser = transformers.HfArgumentParser((ModelArguments, DataArguments, TrainingArguments))
    model_args, data_args, training_args = parser.parse_args_into_dataclasses()

    # model = GOTLlamaForCausalLM.from_pretrained(
    #     model_args.model_name_or_path,
    #     cache_dir=training_args.cache_dir,
    # )

    # tokenizer = transformers.AutoTokenizer.from_pretrained(
    #     model_args.model_name_or_path,
    #     cache_dir=training_args.cache_dir,
    #     model_max_length=training_args.model_max_length,
    #     padding_side="right",
    #     use_fast=False,
    # )

    tokenizer = transformers.AutoTokenizer.from_pretrained("/data/public/ucaswei/cache/Qwen/qwen-chat/", trust_remote_code=True, padding_side="right", model_max_length=training_args.model_max_length,)

    # # model = AutoModelForCausalLM.from_pretrained("/data/public/ucaswei/cache/Qwen/qwen/", device_map="cuda", trust_remote_code=True).eval()

    model = GOTQwenForCausalLM.from_pretrained(model_args.model_name_or_path, low_cpu_mem_usage=True, device_map='cuda')

    smart_tokenizer_and_embedding_resize(
        special_tokens_dict=dict(pad_token=DEFAULT_PAD_TOKEN),
        tokenizer=tokenizer,
        model=model,
        )

    # if data_args.conversation_version == "v0" or "models--decapoda-research--llama-7b-hf" in model_args.model_name_or_path:
    #     if tokenizer.pad_token is None:
    #         smart_tokenizer_and_embedding_resize(
    #             special_tokens_dict=dict(pad_token=DEFAULT_PAD_TOKEN),
    #             tokenizer=tokenizer,
    #             model=model,
    #         )
    #     if "llama" in model_args.model_name_or_path:
    #         tokenizer.add_special_tokens({
    #             "eos_token": DEFAULT_EOS_TOKEN,
    #             "bos_token": DEFAULT_BOS_TOKEN,
    #             "unk_token": DEFAULT_UNK_TOKEN,
    #         })
    # else:
    #     tokenizer.pad_token = tokenizer.unk_token

    dtype = torch.float32
    if training_args.fp16:
        dtype = torch.float16
    if training_args.bf16:
        dtype = torch.bfloat16

    if training_args.lora_enable:
        from peft import LoraConfig, get_peft_model
        lora_config = LoraConfig(
            r=training_args.lora_r,
            lora_alpha=training_args.lora_alpha,
            target_modules=find_all_linear_names(model),
            lora_dropout=training_args.lora_dropout,
            bias=training_args.lora_bias,
            task_type="CAUSAL_LM",
        )
        logging.warning("Adding LoRA adapters...")
        model = get_peft_model(model, lora_config)

    vision_tower_dict = model.get_model().initialize_vision_modules(
        vision_tower=model_args.vision_tower,
        pretrained_stage1_model=model_args.pretrained_stage1_model,
        freeze_vision_tower=model_args.freeze_vision_tower,
        use_im_start_end=model_args.use_im_start_end,
        vision_select_layer=model_args.vision_select_layer,
        dtype=dtype,
        device=training_args.device
    )

    model.initialize_vision_tokenizer(
        tokenizer=tokenizer, 
        freeze_lm_model=model_args.freeze_lm_model, 
        pretrained_stage1_model=model_args.pretrained_stage1_model,
        device=training_args.device,
    )

    model.get_model().vision_tower = create_clip_vit_g(448)
    model.get_model().mm_projector = create_perciever()
    model.to(dtype=dtype, device=training_args.device)

    data_args.image_token_len = vision_tower_dict['image_token_len']
    data_args.image_processor = vision_tower_dict['image_processor']
    data_args.image_processor_high = vision_tower_dict['image_processor_high']
    data_args.use_im_start_end = model_args.use_im_start_end

    # mixed relation, to be fixed
    if model_args.freeze_lm_model:
        model.requires_grad_(False)
        for p in model.get_model().mm_projector.parameters():
            p.requires_grad = True
        for p in model.get_input_embeddings().parameters():
            p.requires_grad = True
        for p in model.get_model().conv_final.parameters():
            p.requires_grad = True
        for p in model.get_model().vision_encoder.parameters():
            p.requires_grad = True

        if not model_args.freeze_vision_tower:
            model.get_model().vision_tower.requires_grad_(True)
            # for i in range(20):
            #     model.get_model().vision_tower.vision_model.encoder.layers[i].requires_grad_(False)
            model.get_model().vision_tower.vision_model.encoder.layers[-1].requires_grad_(False)
            # model.get_model().vision_tower.vision_model.embeddings.requires_grad_(False)
            # model.get_model().vision_tower.vision_model.pre_layrnorm.requires_grad_(False)
            model.get_model().vision_tower.vision_model.post_layernorm.requires_grad_(False)

            for n, p in model.named_parameters():
                print(n, p.requires_grad)
                
    params_grad = [p.numel() for n, p in model.named_parameters() if p.requires_grad]
    print(f"Number of Mapping Trainable Parameters: {sum(params_grad) / (1 << 20):.2f} M")

    # params_no_grad = [n for n, p in model.named_parameters() if not p.requires_grad]
    # if len(params_no_grad) > 0:
    #     if training_args.fsdp is not None and len(training_args.fsdp) > 0:
    #         if len(params_no_grad) < 10:
    #             print('[WARNING] Attempting to use FSDP while {} parameters do not require gradients: {}'. format(len(params_no_grad), params_no_grad))
    #         else:
    #             print('[WARNING] Attempting to use FSDP while {} parameters do not require gradients: {}...(omitted)'. format(len(params_no_grad), ', '.join(params_no_grad[:10])))
    #         print("[WARNING] Attempting to use FSDP with partially frozen paramters, this is experimental.")
    #         print("[WARNING] As of 4/30/23, this feature requires PyTorch-nightly build.  See here for details: https://github.com/haotian-liu/LLaVA#experimental-use-fsdp-to-save-memory-in-pretraining")

    #         from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
    #         def patch_FSDP_use_orig_params(func):
    #             def wrap_func(*args, **kwargs):
    #                 use_orig_params = kwargs.pop('use_orig_params', True)
    #                 return func(*args, **kwargs, use_orig_params=use_orig_params)
    #             return wrap_func

    #         FSDP.__init__ = patch_FSDP_use_orig_params(FSDP.__init__)

    data_module = make_supervised_data_module(
        interleave=training_args.interleave, 
        with_box=training_args.with_box, 
        tokenizer=tokenizer, 
        data_args=data_args
    )

    trainer = GOTTrainer(
        model=model,
        tokenizer=tokenizer,
        args=training_args,
        **data_module)

    if list(pathlib.Path(training_args.output_dir).glob("checkpoint-*")):
        trainer.train(resume_from_checkpoint=True)
    else:
        trainer.train()
    trainer.save_state()

    if training_args.lora_enable:
        state_dict = get_peft_state_maybe_zero_3(
            model.named_parameters(), training_args.lora_bias
        )
        non_lora_state_dict = get_peft_state_non_lora_maybe_zero_3(
            model.named_parameters()
        )
        if training_args.local_rank == 0 or training_args.local_rank == -1:
            model.config.save_pretrained(training_args.output_dir)
            model.save_pretrained(training_args.output_dir, state_dict=state_dict)
            torch.save(non_lora_state_dict, os.path.join(training_args.output_dir, 'non_lora_trainables.bin'))
    else:
        trainer._safe_save(output_dir=training_args.output_dir)


if __name__ == "__main__":
    train()



import modal
import os

achatbot_version = os.getenv("ACHATBOT_VERSION", "0.0.19")
avatar_tag = os.getenv("AVATAR_TAG", "lite_avatar_gpu")
# fastapi_webrtc_bots | fastapi_webrtc_single_bot server
SERVER_TAG = os.getenv("SERVER_TAG", "fastapi_webrtc_single_bot")
image = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg")
    .pip_install(
        [
            "achatbot["
            "fastapi_bot_server,"
            "livekit,livekit-api,daily,agora,"
            "silero_vad_analyzer,"
            "sense_voice_asr,deepgram_asr_processor,"
            "openai_llm_processor,google_llm_processor,litellm_processor,"
            "deep_translator,together_ai,"
            "tts_edge,"
            f"{avatar_tag},"
            "queue"
            f"]=={achatbot_version}",
        ],
        extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
    )
    .pip_install(
        # fix Future exception was never retrieve, when connect timeout ( Connection reset by peer, retry)
        "aiohttp==3.10.11",
        "numpy==1.26.4",
        "torch==2.4.1",
        "torchvision==0.19.1",
        "torchaudio==2.4.1",
    )
    .env(
        {
            "HF_HUB_ENABLE_HF_TRANSFER": "1",
            "ACHATBOT_PKG": "1",
            "LOG_LEVEL": os.getenv("LOG_LEVEL", "info"),
            # asr module engine TAG, default whisper_timestamped_asr
            "ASR_TAG": "sense_voice_asr",
            "ASR_LANG": "zn",
            "ASR_MODEL_NAME_OR_PATH": "/root/.achatbot/models/FunAudioLLM/SenseVoiceSmall",
            # llm processor model, default:google gemini_flash_latest
            "GOOGLE_LLM_MODEL": "gemini-2.0-flash-lite",
            # tts module engine TAG,default tts_edge
            "TTS_TAG": "tts_edge",
        }
    )
)

if avatar_tag == "musetalk_avatar":
    image = (
        image.apt_install("clang")
        .run_commands(
            "which nvcc",
            "which clang++",
            "git clone https://github.com/open-mmlab/mmcv.git",
            "cd /mmcv && git checkout v2.1.0",
            "cd /mmcv && pip install -r requirements/optional.txt",
            "cd /mmcv && FORCE_CUDA=1 TORCH_CUDA_ARCH_LIST='7.5 8.0 8.6 8.7 8.9 9.0' pip install -e . -v",
        )
        .run_commands(
            "pip install -q --no-cache-dir -U openmim",
            "mim install mmengine",
            "mim install 'mmdet==3.3.0'",
            "mim install 'mmpose==1.3.2'",
        )
    )

if avatar_tag == "lam_audio2expression_avatar":
    image = (
        image.pip_install("spleeter==2.4.2")
        .pip_install(
            "typing_extensions==4.14.0",
            "aiortc==1.13.0",
            "protobuf==5.29.4",
            "transformers==4.36.2",
        )
        .env(
            {
                "TRANSPORT": os.getenv("TRANSPORT", "webrtc_websocket_v2"),
                "CONFIG_FILE": os.getenv(
                    "CONFIG_FILE",
                    "/root/.achatbot/config/bots/small_webrtc_fastapi_websocket_avatar_chat_bot.json",
                ),
            }
        )
    )

# image = image.pip_install(
# f"achatbot==0.0.19.post5",
# extra_index_url=os.getenv("EXTRA_INDEX_URL", "https://pypi.org/simple/"),
# )

# ----------------------- app -------------------------------
app = modal.App("fastapi_webrtc_avatar_bot")

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
RESOURCES_DIR = "/root/.achatbot/resources"
resources_vol = modal.Volume.from_name("resources", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_vol = modal.Volume.from_name("assets", create_if_missing=True)
CONFIG_DIR = "/root/.achatbot/config"
config_vol = modal.Volume.from_name("config", create_if_missing=True)
TORCH_CACHE_DIR = "/root/.cache/torch"
torch_cache_vol = modal.Volume.from_name("torch_cache", create_if_missing=True)


# 128 MiB of memory and 0.125 CPU cores by default container runtime
@app.cls(
    image=image,
    gpu=os.getenv("IMAGE_GPU", None),
    secrets=[modal.Secret.from_name("achatbot")],
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        RESOURCES_DIR: resources_vol,
        ASSETS_DIR: assets_vol,
        CONFIG_DIR: config_vol,
        TORCH_CACHE_DIR: torch_cache_vol,
    },
    cpu=4.0,
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=int(os.getenv("IMAGE_MAX_CONTAINERS", "1")),
)
@modal.concurrent(max_inputs=int(os.getenv("IMAGE_CONCURRENT_CN", "10")))  # inputs per container
class Srv:
    @modal.enter()
    def enter(self):
        # run container runtime to enter when container is starting
        import subprocess
        import torch

        subprocess.run("nvidia-smi --version", shell=True)
        gpu_prop = None
        if torch.cuda.is_available():
            gpu_prop = torch.cuda.get_device_properties("cuda:0")
            print(gpu_prop)
            torch.multiprocessing.set_start_method("spawn", force=True)
        else:
            print("CUDA is not available.")

        # todo: init model to load, now use api to load model to run bot with config

    @modal.asgi_app()
    def app(self):
        transport = os.getenv("TRANSPORT", "daily")
        if transport == "webrtc_websocket":
            from achatbot.cmd.webrtc_websocket.fastapi_ws_signaling_bot_serve import (
                app as fastapi_app,
            )

            return fastapi_app
        elif transport == "webrtc_websocket_v2":
            from achatbot.cmd.webrtc_websocket.fastapi_ws_signaling_bot_serve_v2 import (
                app as fastapi_app,
            )

            return fastapi_app
        else:
            SERVER_TAG = os.getenv("SERVER_TAG", "fastapi_webrtc_single_bot")
            if SERVER_TAG == "fastapi_webrtc_single_bot":
                from achatbot.cmd.http.server.fastapi_room_bot_serve import app as fastapi_app

                print("run fastapi_room_bot_serve(single bot)")
                return fastapi_app
            else:
                from achatbot.cmd.http.server.fastapi_daily_bot_serve import app as fastapi_app

                print("run fastapi_daily_bot_serve(multi bots)")
                return fastapi_app


"""
EXTRA_INDEX_URL=https://pypi.org/simple/ ACHATBOT_VERSION=0.0.29 AVATAR_TAG=lam_audio2expression_avatar  IMAGE_CONCURRENT_CN=3  modal serve src/fastapi_webrtc_avatar_bot_serve.py
"""



import os
import logging

from dotenv import load_dotenv

from achatbot.common.factory import EngineClass, EngineFactory
from achatbot.types.codec import CodecArgs
from achatbot.modules.codec.interface import ICodec


load_dotenv(override=True)


class CodecEnvInit:
    @staticmethod
    def getEngine(tag, **kwargs) -> ICodec | EngineClass:
        if "codec_transformers_mimi" == tag:
            from .audio import transformers_mimi
        elif "codec_transformers_dac" == tag:
            from .audio import transformers_dac
        elif "codec_moshi_mimi" == tag:
            from .audio import moshi_mimi
        elif "codec_xcodec2" == tag:
            from .audio import xcodec2
        elif "codec_bitokenizer" == tag:
            from .audio import bicodec
        elif "codec_snac" == tag:
            from .audio import snac
        elif "codec_wavtokenizer" == tag:
            from .audio import wavtokenizer

        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        return engine

    @staticmethod
    def initCodecEngine(tag=None, **kwargs) -> ICodec | EngineClass:
        # codec
        tag = tag or os.getenv("CODEC_TAG", "codec_xcodec2")
        kwargs = kwargs or CodecEnvInit.map_config_func[tag]()
        engine = CodecEnvInit.getEngine(tag, **kwargs)
        logging.info(f"initCodecEngine: {tag}, {engine}")
        return engine

    @staticmethod
    def get_args() -> dict:
        res = CodecArgs(model_dir=os.getenv("CODEC_MODEL_DIR", "")).__dict__
        return res

    # TAG : config
    map_config_func = {
        "codec_wavtokenizer": get_args,
        "codec_snac": get_args,
        "codec_bitokenizer": get_args,
        "codec_xcodec2": get_args,
        "codec_transformers_mimi": get_args,
        "codec_transformers_dac": get_args,
        "codec_moshi_mimi": get_args,
    }
